"""
@author: zeyi-lin
@file: cambricon.py
@time: 2025/03/27 12:32
@description: 寒武纪mlu信息采集
"""

import math
import platform
import subprocess
from typing import Any, Dict, Optional, Tuple

from ..type import HardwareCollector as H
from ..type import HardwareConfig, HardwareFuncResult, HardwareInfoList
from ..utils import generate_key, random_index


def get_cambricon_mlu_info() -> HardwareFuncResult:
    """
    获取寒武纪mlu信息，包括驱动版本、设备信息等
    """
    # cambricon芯片只支持Linux系统
    if platform.system() != "Linux":
        return None, None

    info: Dict[str, Any] = {"driver": None, "mlu": None}
    collector = None
    try:
        driver, mlu_map = map_mlu()
        info["driver"] = driver
        info["mlu"] = mlu_map
        max_mem_value = 0
        for mlu_id in mlu_map:
            mlu_mem = int(mlu_map[mlu_id].get("memory", 0))
            max_mem_value = max(max_mem_value, mlu_mem)
        max_mem_value = max_mem_value * 1024  # 转换为MB
        collector = CambriconCollector(mlu_map, max_mem_value)
    except Exception:  # noqa
        if all(v is None for v in info.values()):
            return None, None
    return info, collector


def map_mlu() -> Tuple[Optional[str], dict]:
    """
    列出所有mlu设备，并包含芯片的映射关系
    返回值样例:
        driver: "1.1.0"
        mlu_map: {"0": { "name": "cambricon-910", "memory": 16, }, "1": { "name": "cambricon-910", "memory": 16, }, ...}
    """
    output = subprocess.run(["cnmon", "info", "-m"], capture_output=True, check=True, text=True).stdout
    mlu_map = {}
    driver = None
    lines = output.split("\n")[1:]

    mlu_id = None
    for line in lines:
        line = line.split()
        try:
            if line[0] == "Card":
                mlu_id = line[-1]
                # 获取mlu的ID
                if mlu_id not in mlu_map:
                    mlu_map[mlu_id] = {}
            if mlu_id is None:
                continue
            # 获取mlu的名称
            if line[0] == "Product":
                # 如果没有mlu_id，则跳过（解析错误）
                mlu_name = line[-1]
                mlu_map[mlu_id]["name"] = mlu_name
            # 获取mlu的驱动版本
            if line[0] == "Driver" and len(mlu_map[mlu_id]) == 1 and driver is None:
                driver = line[-1]
            # 获取mlu的内存大小
            if line[0] == "Total" and len(mlu_map[mlu_id]) == 1:
                memory = line[-2]
                mlu_map[mlu_id]["memory"] = str(int(memory) // 1024)  # 单位为GB
        except Exception:  # noqa
            continue

    return driver, mlu_map


class CambriconCollector(H):
    def __init__(self, mlu_map, max_mem_value):
        super().__init__()
        self.mlu_map = mlu_map
        self.max_mem_value = max_mem_value

        # mlu Utilization (%)
        self.util_key = generate_key("mlu.{mlu_index}.ptc")
        util_config = HardwareConfig(
            y_range=(0, 100),
            chart_index=random_index(),
            chart_name="MLU Utilization (%)",
        )
        self.per_util_configs = {}

        # mlu Memory Allocated (%)
        self.memory_key = generate_key("mlu.{mlu_index}.mem.ptc")
        memory_config = HardwareConfig(
            y_range=(0, 100),
            chart_index=random_index(),
            chart_name="MLU Memory Allocated (%)",
        )
        self.per_memory_configs = {}

        # mlu Memory Allocated (MB)
        self.mem_value_key = generate_key("mlu.{mlu_index}.mem.value")
        mem_value_config = HardwareConfig(
            y_range=(0, self.max_mem_value),
            chart_index=random_index(),
            chart_name="MLU Memory Allocated (MB)",
        )
        self.per_mem_value_configs = {}

        # mlu Temperature (°C)
        self.temp_key = generate_key("mlu.{mlu_index}.temp")
        temp_config = HardwareConfig(
            y_range=(0, None),
            chart_index=random_index(),
            chart_name="MLU Temperature (°C)",
        )
        self.per_temp_configs = {}

        # mlu Power (W)
        self.power_key = generate_key("mlu.{mlu_index}.power")
        power_config = HardwareConfig(
            y_range=(0, None),
            chart_index=random_index(),
            chart_name="MLU Power (W)",
        )
        self.per_power_configs = {}

        for mlu_id in self.mlu_map:
            metric_name = f"MLU {mlu_id}"
            self.per_util_configs[metric_name] = util_config.clone(metric_name=metric_name)
            self.per_memory_configs[metric_name] = memory_config.clone(metric_name=metric_name)
            self.per_mem_value_configs[metric_name] = mem_value_config.clone(metric_name=metric_name)
            self.per_temp_configs[metric_name] = temp_config.clone(metric_name=metric_name)
            self.per_power_configs[metric_name] = power_config.clone(metric_name=metric_name)

    def collect(self) -> HardwareInfoList:
        result: HardwareInfoList = []
        usage_methods = [
            self.get_utilization_usage,
            self.get_memory_usage,
            self.get_mem_value_usage,
            self.get_temperature_usage,
            self.get_power_usage,
        ]

        for method in usage_methods:
            result.extend(method().values())
        return result

    def get_utilization_usage(self) -> dict:
        """
        获取指定mlu设备的利用率
        """
        output = subprocess.run(["cnmon", "info", "-u"], capture_output=True, text=True).stdout

        util_infos = {}
        mlu_ids = []

        # 获取所有mlu ID
        for mlu_id in self.mlu_map:
            mlu_ids.append(mlu_id)

        index = 0
        lines = output.strip().split("\n")
        for line in lines:
            line = line.strip()
            if "mlu average" in line.lower():
                util_infos[mlu_ids[index]] = {
                    "key": self.util_key.format(mlu_index=mlu_ids[index]),
                    "name": f"MLU {mlu_ids[index]} Utilization (%)",
                    "value": math.nan,
                    "config": self.per_util_configs[f"MLU {mlu_ids[index]}"],
                }
                # 获得此mlu的利用率数值
                util = line.split(":")[-1].replace("%", "").strip()
                if util.isdigit():
                    util_infos[mlu_ids[index]]["value"] = float(util)
                index += 1
                continue
        return util_infos

    def get_memory_usage(self) -> dict:
        """
        获取指定mlu设备的显存占用率
        """
        output = subprocess.run(
            ["cnmon", "info", "-m"],
            capture_output=True,
            text=True,
        ).stdout

        memory_infos = {}
        mlu_ids = []

        # 获取所有mlu ID
        for mlu_id in self.mlu_map:
            mlu_ids.append(mlu_id)

        index = 0
        lines = output.strip().split("\n")
        for line_idx, line in enumerate(lines):
            # 如果包含mlu average，则表示该行是mlu的利用率
            if "physical memory usage" in line.lower():
                if "used" in lines[line_idx + 2].lower():
                    used_line = lines[line_idx + 2]
                    # 初始化mlu的利用率
                    memory_infos[mlu_ids[index]] = {
                        "key": self.memory_key.format(mlu_index=mlu_ids[index]),
                        "name": f"MLU {mlu_ids[index]} Memory Allocated (%)",
                        "value": math.nan,
                        "config": self.per_memory_configs[f"MLU {mlu_ids[index]}"],
                    }
                    # 获得此mlu的显存数值（MiB）
                    memory = used_line.split(":")[-1].replace("MiB", "").strip()
                    if memory.isdigit():
                        # 计算mlu显存占用率
                        memory_infos[mlu_ids[index]]["value"] = (
                            float(memory) / (float(self.mlu_map[mlu_ids[index]]["memory"]) * 1024) * 100
                        )
                    index += 1
                    continue
        return memory_infos

    def get_mem_value_usage(self) -> dict:
        """
        获取指定mlu设备的显存使用量 (MB)
        """
        output = subprocess.run(
            ["cnmon", "info", "-m"],
            capture_output=True,
            text=True,
        ).stdout

        mem_value_infos = {}
        mlu_ids = []

        # 获取所有mlu ID
        for mlu_id in self.mlu_map:
            mlu_ids.append(mlu_id)

        index = 0
        lines = output.strip().split("\n")
        for line_idx, line in enumerate(lines):
            # 如果包含mlu average，则表示该行是mlu的利用率
            if "physical memory usage" in line.lower():
                if "used" in lines[line_idx + 2].lower():
                    used_line = lines[line_idx + 2]
                    # 初始化mlu的利用率
                    mem_value_infos[mlu_ids[index]] = {
                        "key": self.mem_value_key.format(mlu_index=mlu_ids[index]),
                        "name": f"MLU {mlu_ids[index]} Memory Allocated (MB)",
                        "value": math.nan,
                        "config": self.per_mem_value_configs[f"MLU {mlu_ids[index]}"],
                    }
                    # 获得此mlu的显存数值（MiB）
                    memory = used_line.split(":")[-1].replace("MiB", "").strip()
                    if memory.isdigit():
                        mem_value_infos[mlu_ids[index]]["value"] = float(memory)
                    index += 1
                    continue
        return mem_value_infos

    def get_temperature_usage(self) -> dict:
        """
        获取指定mlu设备的温度(°C)
        """
        output = subprocess.run(["cnmon", "info", "-e"], capture_output=True, text=True).stdout
        temp_infos = {}
        mlu_ids = []

        # 获取所有mlu ID
        for mlu_id in self.mlu_map:
            mlu_ids.append(mlu_id)

        index = 0
        lines = output.strip().split("\n")
        for line_idx, line in enumerate(lines):
            if "temperature" in line.lower():
                if "chip" in lines[line_idx + 2].lower():
                    temp_line = lines[line_idx + 2]
                    temp_infos[mlu_ids[index]] = {
                        "key": self.temp_key.format(mlu_index=mlu_ids[index]),
                        "name": f"MLU {mlu_ids[index]} Temperature (°C)",
                        "value": math.nan,
                        "config": self.per_temp_configs[f"MLU {mlu_ids[index]}"],
                    }
                    # 获得此mlu的温度数值
                    temp = temp_line.split(":")[-1].replace("C", "").strip()
                    if temp.isdigit():
                        temp_infos[mlu_ids[index]]["value"] = float(temp)
                    index += 1
                    continue
        return temp_infos

    def get_power_usage(self) -> dict:
        """
        获取指定mlu设备的功耗(W)
        """
        output = subprocess.run(["cnmon", "info", "-p"], capture_output=True, text=True).stdout
        power_infos = {}
        mlu_ids = []

        # 获取所有mlu ID
        for mlu_id in self.mlu_map:
            mlu_ids.append(mlu_id)

        index = 0
        lines = output.strip().split("\n")
        for line_idx, line in enumerate(lines):
            if "power" in line.lower() and line_idx + 1 < len(lines):
                if "usage" in lines[line_idx + 1].lower():
                    power_line = lines[line_idx + 1]
                    power_infos[mlu_ids[index]] = {
                        "key": self.power_key.format(mlu_index=mlu_ids[index]),
                        "name": f"MLU {mlu_ids[index]} Power (W)",
                        "value": math.nan,
                        "config": self.per_power_configs[f"MLU {mlu_ids[index]}"],
                    }
                    # 获得此mlu的功耗数值
                    power = power_line.split(":")[-1].replace("W", "").strip()
                    if power.isdigit():
                        power_infos[mlu_ids[index]]["value"] = float(power)
                    index += 1
                    continue
        return power_infos



"""SwanLab Object3D Module

This module provides classes for handling 3D data types like point clouds and meshes.

Classes:
    Object3D: Main dispatcher class for handling different types of 3D data
    PointCloud: Class for handling point cloud data with XYZ, XYZC, and XYZRGB formats
    Model3D: Class for handling 3D model files like GLB
    Molecule: Class for handling molecule data from various formats by RDKit

Examples:
    # Create point cloud from XYZ coordinates
    >>> import numpy as np
    >>> from swanlab.data.modules.object3d import Object3D
    >>> points = np.random.rand(100, 3)  # XYZ format
    >>> obj = Object3D(points)  # Returns PointCloud object

    # Create point cloud from XYZC data (coordinates + category)
    >>> points_c = np.random.rand(100, 4)  # XYZC format
    >>> obj = Object3D(points_c)  # Category colors mapped automatically

    # Create point cloud from XYZRGB data
    >>> points_rgb = np.random.rand(100, 6)  # XYZRGB format
    >>> obj = Object3D(points_rgb, step=1, caption="My Points")

    # Load from files
    >>> obj = Object3D("model.glb")  # Load 3D model
    >>> obj = Object3D("cloud.swanlab.pts.json")  # Load point cloud

    # Direct class usage
    >>> from swanlab.data.modules.object3d import PointCloud, Model3D
    >>> pc = PointCloud.from_xyz(points, step=1)
    >>> model = Model3D.from_glb_file("model.glb", caption="My Model")
"""

from .model3d import Model3D
from .molecule import Molecule
from .object3d import Object3D
from .point_cloud import PointCloud

__all__ = [
    'Object3D',
    'PointCloud',
    'Model3D',
    'Molecule',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/10 20:08
@File: test_model.py
@IDE: pycharm
@Description:
    测试上传的模型类型
"""

import random
import time

from nanoid import generate

from swanlab.core_python.uploader import FileModel


class TestFileModel:

    def test_create(self):
        """
        测试FileModel的create函数
        """
        # 生成一系列的FileModel
        file_models = []
        lr = None  # 最后一个requirements, str
        lm = None  # 最后一个metadata, dict
        lc = None  # 最后一个config, dict
        lo = None  # 最后一个conda str
        for i in range(10):
            lr = generate() if random.random() > 0.5 else lr
            lm = {generate(): generate()} if random.random() > 0.5 else lm
            lc = {generate(): generate()} if random.random() > 0.5 else lc
            lo = generate() if random.random() > 0.5 else lo
            file_models.append(FileModel(lr, lm, lc, lo))
            time.sleep(0.01)
        # 验证
        # file_models打乱顺序
        random.shuffle(file_models)
        file_model = FileModel.create(file_models)
        assert file_model.requirements == lr
        assert file_model.metadata == lm
        assert file_model.config == lc
        assert file_model.conda == lo

    def test_empty(self):
        assert FileModel().empty is True
        assert FileModel("1", None, None, None).empty is False
        assert FileModel(None, {}, None, None).empty is False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-04-26 17:07:44
@File: swanlab\integration\zhipuai\resolver.py
@IDE: vscode
@Description:
    ZhipuAI Resolver
"""
import datetime
import io
import os
import sys
import swanlab
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Sequence

current_dir = os.path.dirname(os.path.abspath(__file__))
relative_path = os.path.join(current_dir, "..")
sys.path.append(relative_path)

from integration_utils.autologging import Response


@dataclass
class UsageMetrics:
    elapsed_time: float = None
    prompt_tokens: int = None
    completion_tokens: int = None
    total_tokens: int = None


@dataclass
class Metrics:
    usage: UsageMetrics = None
    stats: swanlab.Text = None


class ZhipuAIClientResponseResolver:
    def __init__(self):
        self.define_metrics_called = False

    def __call__(
        self,
        args: Sequence[Any],
        kwargs: Dict[str, Any],
        response: Response,
        lib_version: str,
        start_time: float,  # pass to comply with the protocol, but use response["created"] instead
        time_elapsed: float,
    ) -> Optional[Dict[str, Any]]:
        request = kwargs
        lib_version = lib_version

        try:

            response = self.chat_completion_to_dict(response)
            return self._resolve_chat_completion(request, response, lib_version, time_elapsed)

        except Exception as e:
            print(f"Failed to resolve request/response: {e}")
        return None

    def _resolve_chat_completion(
        self,
        request: Dict[str, Any],
        response: Response,
        lib_version: str,
        time_elapsed: float,
    ) -> Dict[str, Any]:
        """Resolves the request and response objects for `zhipuai.ZhipuAI().Completion`."""
        prompt = io.StringIO()
        for message in request["messages"]:
            prompt.write(f"\n\n**{message['role']}**: {message['content']}\n")
        request_str = prompt.getvalue()
        response_choices = response.get("choices")
        choices = [
            f"\n\n**{choice['message']['role']}**: {choice['message']['content']}\n" for choice in response_choices
        ]

        return self._resolve_metrics(
            request=request,
            response=response,
            lib_version=lib_version,
            request_str=request_str,
            choices=choices,
            time_elapsed=time_elapsed,
        )

    def _resolve_metrics(
        self,
        request: Dict[str, Any],
        response: Response,
        lib_version: str,
        request_str: str,
        choices: List[str],
        time_elapsed: float,
    ) -> Dict[str, Any]:
        """Resolves the request and response objects for `zhipuai.ZhipuAI().Completion`."""
        results = [{"inputs": {"request": request_str}, "outputs": {"response": choice}} for choice in choices]
        metrics = self._get_metrics_to_log(request, response, lib_version, results, time_elapsed)
        return self._convert_metrics_to_dict(metrics)

    @staticmethod
    def _get_usage_metrics(response: Response, time_elapsed: float) -> UsageMetrics:
        """Gets the usage stats from the response object."""
        if response.get("usage"):
            usage_stats = UsageMetrics(**response["usage"])
        else:
            usage_stats = UsageMetrics()
        usage_stats.elapsed_time = time_elapsed
        return usage_stats

    def _get_metrics_to_log(
        self,
        request: Dict[str, Any],
        response: Response,
        lib_version: str,
        results: List[Any],
        time_elapsed: float,
    ) -> Metrics:
        model = response.get("model")
        usage_metrics = self._get_usage_metrics(response, time_elapsed)

        usage = []

        for result in results:
            swanlab.log(
                {
                    "result": [
                        swanlab.Text(caption="request", data=str(request["messages"])),
                        swanlab.Text(caption="response", data=response["choices"][0]["message"]["content"]),
                        swanlab.Text(caption="lib_version", data=lib_version),
                        swanlab.Text(caption="request_id", data=response["id"]),
                        swanlab.Text(caption="model", data=response["model"]),
                        swanlab.Text(caption="completion_tokens", data=response["usage"]["completion_tokens"]),
                        swanlab.Text(caption="prompt_tokens", data=response["usage"]["prompt_tokens"]),
                        swanlab.Text(caption="total_tokens", data=response["usage"]["total_tokens"]),
                        swanlab.Text(
                            caption="start_time", data=str(datetime.datetime.fromtimestamp(response["created"]))
                        ),
                        swanlab.Text(
                            caption="end_time",
                            data=str(datetime.datetime.fromtimestamp(response["created"] + time_elapsed)),
                        ),
                        swanlab.Text(caption="used_time", data=time_elapsed),
                    ]
                }
            )

            row = {
                "request": str(request["messages"]),
                "response": response["choices"][0]["message"]["content"],
                "lib_version": lib_version,
                "model": model,
                "start_time": datetime.datetime.fromtimestamp(response["created"]),
                "end_time": datetime.datetime.fromtimestamp(response["created"] + time_elapsed),
                "request_id": response.get("id", None),
            }

        result_text = swanlab.Text(caption="result_row_dict", data=str(row))

        metrics = Metrics(stats=result_text, usage=usage_metrics)
        return metrics

    @staticmethod
    def _convert_metrics_to_dict(metrics: Metrics) -> Dict[str, Any]:
        """Converts metrics to a dict."""
        metrics_dict = {
            "stats": metrics.stats,
        }
        usage_stats = {f"usage/{k}": v for k, v in asdict(metrics.usage).items()}
        metrics_dict.update(usage_stats)
        return metrics_dict

    def chat_completion_to_dict(self, completion):
        data = {
            "model": completion.model,
            "created": completion.created,
            "choices": [
                {
                    "index": choice.index,
                    "finish_reason": choice.finish_reason,
                    "message": {
                        "content": choice.message.content,
                        "role": choice.message.role,
                        "tool_calls": choice.message.tool_calls,
                    },
                }
                for choice in completion.choices
            ],
            "request_id": completion.request_id,
            "id": completion.id,
            "usage": {
                "prompt_tokens": completion.usage.prompt_tokens,
                "completion_tokens": completion.usage.completion_tokens,
                "total_tokens": completion.usage.total_tokens,
            },
        }
        return data



from io import BytesIO
from typing import Union, Any, TYPE_CHECKING

from swanlab.toolkit import MediaBuffer, DataSuite as D, MediaType

if TYPE_CHECKING:
    # noinspection PyPackageRequirements
    import matplotlib as _matplotlib  # type: ignore

    # noinspection PyPackageRequirements
    import numpy as _numpy  # type: ignore

    # noinspection PyPackageRequirements
    import torch as _torch  # type: ignore

    # noinspection PyPackageRequirements
    from PIL.Image import Image as _PILImage  # type: ignore

    TorchTensorType = Union["_torch.Tensor", "_torch.Variable"]
    ImageDataType = Union["_matplotlib.artist.Artist", "_PILImage", "TorchTensorType", "_numpy.ndarray"]
    ImageDataOrPathType = Union[str, "_PILImage", ImageDataType]


def is_pytorch_tensor_typename(typename: str) -> bool:
    return typename.startswith("torch.") and ("Tensor" in typename or "Variable" in typename)


def get_full_typename(o: Any) -> Any:
    """Determine types based on type names.

    Avoids needing to import (and therefore depend on) PyTorch, TensorFlow, etc.
    """
    instance_name = o.__class__.__module__ + "." + o.__class__.__name__
    if instance_name in ["builtins.module", "__builtin__.module"]:
        return o.__name__
    else:
        return instance_name


def convert_size(size=None):
    """将size转换为PIL图像的size"""
    if size is None:
        return None

    elif isinstance(size, int):
        return size

    elif isinstance(size, (list, tuple)) and len(size) in [1, 2]:
        size = tuple(size)
        width = int(size[0]) if size[0] is not None else None
        height = int(size[1]) if len(size) > 1 and size[1] is not None else None

        return (width, height) if len(size) == 2 else width

    raise ValueError("swanlab.Image - param `size` must be a list (with 2 or 1 elements) or an int")


class Image(MediaType):
    ACCEPT_FORMAT = ["png", "jpg", "jpeg", "bmp", "gif"]

    def __init__(
        self,
        data_or_path: "ImageDataOrPathType",
        mode: str = None,
        caption: str = None,
        file_type: str = None,
        size: Union[int, list, tuple] = None,
    ):
        """Image class constructor

        Parameters
        ----------
        data_or_path: (str or numpy.ndarray or PIL.Image or torch.Tensor or matplotlib figure)
            Path to the image file, numpy array of image data, PIL.Image data, torch.Tensor data or
            matplotlib figure data.
        mode: (str)
            The PIL Mode for an image. Most common is 'L', 'RGB', 'RGBA'.
            More information about the mode can be found at
            https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes
        caption: (str)
            Caption for the image.
        file_type: (str)
            File type for the image. It is used to save the image in the specified format. The default is 'png'.
            The supported file types are ['png', 'jpg', 'jpeg', 'bmp', 'gif'].
        size: (int or list or tuple)
            The size of the image can be controlled in four ways:
            If int type, it represents the maximum side length of the image, that is,
            the width and height cannot exceed this maximum side length. The image will be scaled proportionally to
            ensure that the maximum side length does not exceed MAX_DIMENSION;

            If list or tuple type with both specified values, e.g. (500, 500),
            then the image will be scaled to the specified width and height;

            If list or tuple type with only one specified value and another value as None, e.g. (500, None),
            it means resize the image to the specified width, and the height is scaled proportionally.

            If it is None, it means no scaling for the image.
        """

        try:
            # noinspection PyPackageRequirements
            from PIL import Image as PILImage

            # noinspection PyPackageRequirements
            import numpy as np
        except ImportError:
            raise ImportError(
                "pillow and numpy are required for Image class, you can install them by `pip install pillow numpy`"
            )

        super().__init__()
        self.format = self.__convert_file_type(file_type)
        self.size = convert_size(size)

        # 如果输入为路径字符串
        if isinstance(data_or_path, str):
            # 如果文件后缀为gif，则将format设置为gif
            if data_or_path.endswith(".gif"):
                self.format = "gif"
                # 直接保存到buffer中
                with open(data_or_path, "rb") as f:
                    self.buffer = MediaBuffer()
                    self.buffer.write(f.read())
                self.caption = D.check_caption(caption)
                return
            
            # 否则，使用PILImage打开图像
            try:
                image_data = PILImage.open(data_or_path).convert(mode)
            except Exception as e:
                raise ValueError(f"Invalid image path: {data_or_path}") from e
        elif isinstance(data_or_path, PILImage.Image):
            # 如果输入为PIL.Image
            image_data = data_or_path.convert(mode)
        elif is_pytorch_tensor_typename(get_full_typename(data_or_path)):
            # 如果输入为pytorch tensor
            try:
                import torchvision  # noqa
            except ImportError:
                raise TypeError(
                    "swanlab.Image requires `torchvision` when process torch.tensor data. "
                    "Install with 'pip install torchvision'."
                )
            if hasattr(data_or_path, "requires_grad") and data_or_path.requires_grad:
                data_or_path = data_or_path.detach()  # noqa
            if hasattr(data_or_path, "detype") and str(data_or_path.type) == "torch.uint8":
                data_or_path = data_or_path.to(float)  # type: ignore
            data_or_path = torchvision.utils.make_grid(data_or_path, normalize=True)
            image_data = PILImage.fromarray(
                data_or_path.mul(255).clamp(0, 255).byte().permute(1, 2, 0).cpu().numpy(), mode=mode
            )
        elif hasattr(data_or_path, "savefig"):
            # 如果输入为matplotlib图像
            try:
                # noinspection PyPackageRequirements
                import matplotlib
            except ImportError:
                raise ImportError(
                    "swanlab.Image requires `matplotlib` when process matplotlib.artist.Artist data. "
                    "you can install them by `pip install matplotlib`"
                )

            try:
                buf = BytesIO()
                data_or_path.savefig(buf, format=self.format)  # 将图像保存到BytesIO对象
                buf.seek(0)  # 移动到缓冲区的开始位置
                image_data = PILImage.open(buf).convert(mode)  # 使用PIL打开图像
                buf.close()  # 关闭缓冲区
            except Exception as e:
                raise TypeError("Invalid matplotlib figure for the image") from e

        elif isinstance(data_or_path, np.ndarray):
            # 如果输入为numpy array
            try:
                if data_or_path.ndim == 2 or (data_or_path.ndim == 3 and data_or_path.shape[2] in [3, 4]):
                    image_data = PILImage.fromarray(np.clip(data_or_path, 0, 255).astype(np.uint8), mode=mode)
                else:
                    raise TypeError("Invalid numpy array: the numpy array must be 2D or 3D with 3 or 4 channels.")
            except Exception as e:
                raise TypeError("Invalid numpy array for the image") from e

        else:
            # 以上都不是，则报错
            raise TypeError(
                "Unsupported image type. Please provide a valid path, numpy array, PIL.Image, torch."
                "Tensor or matplotlib figure."
            )

        self.image_data = self.__resize(image_data, self.size)
        """
        转换为矩阵后的数据
        """
        self.buffer = MediaBuffer()
        self.image_data.save(self.buffer, format=self.format if self.format != "jpg" else "jpeg")
        self.caption = D.check_caption(caption)

    def __convert_file_type(self, file_type: str = None):
        """转换file_type，并检测file_type是否正确"""
        if file_type is None:
            file_type = "png"

        if file_type not in self.ACCEPT_FORMAT:
            raise ValueError(f"file_type must be one of {self.ACCEPT_FORMAT}")

        return file_type

    def __resize(self, image, size=None):
        """将图像调整大小"""
        if size is None:
            self.size = image.size
            return image

        elif isinstance(size, int):
            if max(image.size) > size:
                image.thumbnail((size, size))
            return image

        elif isinstance(size, (list, tuple)):
            size = tuple(size)
            if all(size):
                return image.resize(size)
            if size[0] is not None:
                w_percent = size[0] / float(image.size[0])
                hsize = int(float(image.size[1]) * w_percent)
                return image.resize((size[0], hsize))
            if size[1] is not None:
                h_percent = size[1] / float(image.size[1])
                w_size = int(float(image.size[0]) * h_percent)
                return image.resize((w_size, size[1]))

        raise ValueError("swanlab.Image - param `size` must be a list (with 2 or 1 elements) or an int")

    @property
    def image_size(self):
        """
        得到图像的大小
        """
        return self.image_data.size

    # ---------------------------------- 覆写方法 ----------------------------------

    def parse(self):
        # 文件名称
        if self.format == "gif":
            hash_name = D.get_hash_by_bytes(self.buffer.getvalue())[:16]
        else:
            hash_name = D.get_hash_by_pil(self.image_data)[:16]
        save_name = f"image-step{self.step}-{hash_name}.{self.format}"
        return save_name, self.buffer

    def get_more(self):
        """返回more数据"""
        return {"caption": self.caption} if self.caption is not None else None

    def get_section(self) -> str:
        """设定分组名"""
        return "Image"

    def get_chart(self):
        return self.Chart.IMAGE



"""
@author: cunyue
@file: nvidia.py
@time: 2024/12/3 20:12
@description: NVIDIA GPU信息采集
"""

import subprocess

import pynvml

from swanlab.log import swanlog
from ..type import HardwareFuncResult, HardwareCollector, HardwareInfoList, HardwareInfo
from ..utils import generate_key, HardwareConfig, random_index


def get_nvidia_gpu_info() -> HardwareFuncResult:
    """获取 GPU 信息"""

    info = {"driver": None, "cores": None, "type": [], "memory": [], "cuda": None, "architecture": [], "cudacores": []}
    max_gpu_mem_mb = 0
    try:
        pynvml.nvmlInit()
    except Exception:  # noqa
        return None, None
    try:
        # 获取 NVIDIA 驱动版本信息
        nv_driver = pynvml.nvmlSystemGetDriverVersion()
        if isinstance(nv_driver, bytes):
            nv_driver = nv_driver.decode("utf-8")
        info["driver"] = nv_driver

        # 获取 CUDA 版本
        info["cuda"] = get_cuda_version()

        # 获取 NVIDIA GPU 数量
        info["cores"] = pynvml.nvmlDeviceGetCount()
        # 遍历每个 GPU，获取 GPU 信息
        for i in range(info["cores"]):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            # 获取 GPU 型号
            gpu_name = pynvml.nvmlDeviceGetName(handle)  # types: bytes | str
            if isinstance(gpu_name, bytes):  # Fix for pynvml 早期版本，关联 issue: #605
                gpu_name = gpu_name.decode("utf-8")
            info["type"].append(gpu_name)
            # 获取 GPU 的总显存
            total_memory = pynvml.nvmlDeviceGetMemoryInfo(handle).total >> 20  # MB
            max_gpu_mem_mb = max(max_gpu_mem_mb, total_memory)
            info["memory"].append(str(round(total_memory / 1024)))  # GB
            # 获取 GPU 架构
            try:
                NVIDIA_GPU_ARCHITECTURE = {
                    pynvml.NVML_DEVICE_ARCH_KEPLER: "Kepler",  # example: GeForce GTX 680, GeForce GTX 780, Tesla K80
                    pynvml.NVML_DEVICE_ARCH_MAXWELL: "Maxwell",  # example: GeForce GTX 750 Ti, GeForce GTX 980, Tesla M40
                    pynvml.NVML_DEVICE_ARCH_PASCAL: "Pascal",  # example: GeForce GTX 1080 Ti, GeForce GTX 1060, Tesla P100
                    pynvml.NVML_DEVICE_ARCH_VOLTA: "Volta",  # example: Tesla V100, Titan V
                    pynvml.NVML_DEVICE_ARCH_TURING: "Turing",  # example: GeForce RTX 2080 Ti, GeForce GTX 1660 Ti, Tesla T4
                    pynvml.NVML_DEVICE_ARCH_AMPERE: "Ampere",  # example: GeForce RTX 3080, GeForce RTX 3060, A100
                    pynvml.NVML_DEVICE_ARCH_ADA: "Ada",  # example: GeForce RTX 4090, GeForce RTX 4080, L40
                    pynvml.NVML_DEVICE_ARCH_HOPPER: "Hopper",  # example: H100, H800
                    pynvml.NVML_DEVICE_ARCH_UNKNOWN: "Unknown",
                }
                info["architecture"].append(NVIDIA_GPU_ARCHITECTURE[pynvml.nvmlDeviceGetArchitecture(handle)])
            except Exception:
                pass
            # 获取 GPU 的CUDA核心数
            try:
                info["cudacores"].append(pynvml.nvmlDeviceGetNumGpuCores(handle))
            except Exception:
                pass

    except UnicodeDecodeError:  # 部分GPU型号无法解码
        return None, None
    except pynvml.NVMLError:
        pass
    finally:
        pynvml.nvmlShutdown()
        count = info["cores"]
        return info, None if not count else GpuCollector(count=count, max_mem_mb=max_gpu_mem_mb)


def get_cuda_version():
    """获取 CUDA 版本"""
    try:
        output = subprocess.check_output(["nvcc", "--version"]).decode("utf-8")
        for line in output.split('\n'):
            if "release" in line:
                version = line.split("release")[-1].strip().split(" ")[0][:-1]
                return version
    except Exception:  # noqa
        return None


class GpuCollector(HardwareCollector):

    def __init__(self, count: int, max_mem_mb: int):
        super().__init__()
        # GPU 利用率
        self.gpu_util_key = generate_key("gpu.{idx}.pct")
        util_config = HardwareConfig(y_range=(0, 100), chart_name="GPU Utilization (%)", chart_index=random_index())
        # GPU 内存使用率
        self.gpu_mem_pct_key = generate_key("gpu.{idx}.mem.pct")
        mem_pct_config = HardwareConfig(
            y_range=(0, 100), chart_name="GPU Memory Allocated (%)", chart_index=random_index()
        )
        # GPU 内存使用量
        self.gpu_mem_value_key = generate_key("gpu.{idx}.mem.value")
        mem_value_config = HardwareConfig(
            y_range=(0, max_mem_mb), chart_name="GPU Memory Allocated (MB)", chart_index=random_index()
        )
        # GPU 温度
        self.gpu_temp_key = generate_key("gpu.{idx}.temp")
        tem_config = HardwareConfig(chart_name="GPU Temperature (℃)", chart_index=random_index())
        # GPU 功耗
        self.gpu_power_key = generate_key("gpu.{idx}.power")
        power_config = HardwareConfig(chart_name="GPU Power Usage (W)", chart_index=random_index())
        # GPU 访存时间百分比
        self.gpu_mem_time_key = generate_key("gpu.{idx}.mem.time")
        mem_time_config = HardwareConfig(
            y_range=(0, 100), chart_name="GPU Time Spent Accessing Memory (%)", chart_index=random_index()
        )
        # 每个GPU的配置信息
        self.per_gpu_configs = {
            self.gpu_mem_pct_key: [],
            self.gpu_mem_value_key: [],
            self.gpu_temp_key: [],
            self.gpu_power_key: [],
            self.gpu_mem_time_key: [],
            self.gpu_util_key: [],
        }
        self.handles = []
        for idx in range(count):
            metric_name = "GPU {idx}".format(idx=idx)
            self.per_gpu_configs[self.gpu_mem_pct_key].append(mem_pct_config.clone(metric_name=metric_name))
            self.per_gpu_configs[self.gpu_mem_value_key].append(mem_value_config.clone(metric_name=metric_name))
            self.per_gpu_configs[self.gpu_temp_key].append(tem_config.clone(metric_name=metric_name))
            self.per_gpu_configs[self.gpu_power_key].append(power_config.clone(metric_name=metric_name))
            self.per_gpu_configs[self.gpu_mem_time_key].append(mem_time_config.clone(metric_name=metric_name))
            self.per_gpu_configs[self.gpu_util_key].append(util_config.clone(metric_name=metric_name))

    @HardwareCollector.try_run()
    def get_gpu_config(self, key: str, idx: int) -> HardwareConfig:
        """
        获取 某个GPU的某个配置信息
        """
        return self.per_gpu_configs[key][idx]

    @HardwareCollector.try_run()
    def get_gpu_util(self, idx: int) -> HardwareInfo:
        """
        获取 GPU 利用率
        """
        handle = self.handles[idx]
        util_info = pynvml.nvmlDeviceGetUtilizationRates(handle)
        return {
            "key": self.gpu_util_key.format(idx=idx),
            "value": util_info.gpu,
            "name": "GPU {idx} Utilization (%)".format(idx=idx),
            "config": self.get_gpu_config(self.gpu_util_key, idx),
        }

    @HardwareCollector.try_run()
    def get_gpu_mem_pct(self, idx: int) -> HardwareInfo:
        """
        获取 GPU 内存使用率
        """
        handle = self.handles[idx]
        mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        mem_pct = mem_info.used / mem_info.total * 100
        return {
            "key": self.gpu_mem_pct_key.format(idx=idx),
            "value": mem_pct,
            "name": "GPU {idx} Memory Allocated (%)".format(idx=idx),
            "config": self.get_gpu_config(self.gpu_mem_pct_key, idx),
        }

    @HardwareCollector.try_run()
    def get_gpu_mem_value(self, idx: int) -> HardwareInfo:
        """
        获取 GPU 内存使用量(MB)
        """
        handle = self.handles[idx]
        mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        return {
            "key": self.gpu_mem_value_key.format(idx=idx),
            "value": mem_info.used >> 20,
            "name": "GPU {idx} Memory Allocated (MB)".format(idx=idx),
            "config": self.get_gpu_config(self.gpu_mem_value_key, idx),
        }

    @HardwareCollector.try_run()
    def get_gpu_temp(self, idx: int) -> HardwareInfo:
        """
        获取 GPU 温度
        """
        handle = self.handles[idx]
        temp_info = pynvml.nvmlDeviceGetTemperature(handle, pynvml.NVML_TEMPERATURE_GPU)
        return {
            "key": self.gpu_temp_key.format(idx=idx),
            "value": temp_info,
            "name": "GPU {idx} Temperature (℃)".format(idx=idx),
            "config": self.get_gpu_config(self.gpu_temp_key, idx),
        }

    @HardwareCollector.try_run()
    def get_gpu_power(self, idx: int) -> HardwareInfo:
        """
        获取 GPU 功耗
        """
        handle = self.handles[idx]
        # 功耗单位为mW，转换为W
        power_info = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000
        return {
            "key": self.gpu_power_key.format(idx=idx),
            "value": power_info,
            "name": "GPU {idx} Power Usage (W)".format(idx=idx),
            "config": self.get_gpu_config(self.gpu_power_key, idx),
        }

    @HardwareCollector.try_run()
    def get_gpu_mem_time(self, idx: int) -> HardwareInfo:
        """
        获取 GPU 访存时间百分比
        """
        handle = self.handles[idx]
        info = pynvml.nvmlDeviceGetUtilizationRates(handle)
        return {
            "key": self.gpu_mem_time_key.format(idx=idx),
            "value": info.memory,
            "name": "GPU {idx} Time Spent Accessing Memory (%)".format(idx=idx),
            "config": self.get_gpu_config(self.gpu_mem_time_key, idx),
        }

    def collect(self) -> HardwareInfoList:
        """
        采集信息
        """
        result: HardwareInfoList = []
        for idx, handle in enumerate(self.handles):
            result.append(self.get_gpu_mem_pct(idx))
            result.append(self.get_gpu_mem_value(idx))
            result.append(self.get_gpu_util(idx))
            result.append(self.get_gpu_temp(idx))
            result.append(self.get_gpu_power(idx))
            result.append(self.get_gpu_mem_time(idx))
        return result

    def __del__(self):
        pynvml.nvmlShutdown()

    def before_collect_impl(self):
        # 低频采集下（30s以下），应该每次采集时都执行pynvml.nvmlInit()
        # 高频采集下（30s以上），应该在初始化时执行pynvml.nvmlInit()，在最后一次采集时执行pynvml.nvmlShutdown()
        # 在外部定时任务处，超过10次即变为低频采集，因此需要判断一下
        pynvml.nvmlInit()
        for i in range(pynvml.nvmlDeviceGetCount()):
            self.handles.append(pynvml.nvmlDeviceGetHandleByIndex(i))
        swanlog.debug("NVIDIA GPU nvml inited.")

    def after_collect_impl(self):
        pynvml.nvmlShutdown()
        self.handles.clear()
        swanlog.debug("NVIDIA GPU nvml shutdown.")



"""
// @author: ComPleHN
// @file: graphgl.vue
// @time: 2025/5/27 16:29
// @description: 本文件是对于echarts的 GL关系图 图表测试
"""
# ---------------------------------------------- Graphgl - Basic_graphgl ----------------------------------------------
import random

from pyecharts import options as opts
from pyecharts.charts import GraphGL


nodes = []
for i in range(50):
    for j in range(50):
        nodes.append(
            opts.GraphGLNode(
                x=random.random() * 958,
                y=random.random() * 777,
                value=1,
            )
        )

links = []
for i in range(50):
    for j in range(50):
        if i < 50 - 1:
            links.append(
                opts.GraphGLLink(
                    source=i + j * 50,
                    target=i + 1 + j * 50,
                    value=1,
                )
            )
        if j < 50 - 1:
            links.append(
                opts.GraphGLLink(
                    source=i + j * 50,
                    target=i + (j + 1) * 50,
                    value=1,
                )
            )

c1 = (
    GraphGL(init_opts=opts.InitOpts())
    .add(
        series_name="",
        nodes=nodes,
        links=links,
        itemstyle_opts=opts.ItemStyleOpts(color="rgba(255,255,255,0.8)"),
        linestyle_opts=opts.LineStyleOpts(color="rgba(255,255,255,0.8)", width=3),
        force_atlas2_opts=opts.GraphGLForceAtlas2Opts(
            steps=5,
            edge_weight_influence=4,
        ),
    )
    .set_dark_mode()
)

import swanlab

swanlab.init(
    project="echarts-test",
    experiment_name="graphgl",
    public=True,
)

swanlab.log(
    {
        "Graphgl - Basic_graphgl": c1,
    }
)



"""
@author: cunyue
@file: store.py
@time: 2025/6/20 17:00
@description: 运行时配置
NOTE: 只允许在 swanlab/data 模块下访问，其他地方不允许访问
"""

import functools
import inspect
import os.path
from typing import Optional, List, Literal, Dict, Tuple

from pydantic import BaseModel

# 定义远程指标类型
RemoteMetric = Dict[str, Tuple[str, str, Optional[dict], Optional[int]]]


class RunStore(BaseModel):

    # ---------------------------------- 项目 ----------------------------------
    # 项目名称
    project: Optional[str] = None
    # 项目所在空间
    workspace: Optional[str] = None
    # 项目可见性
    visibility: Optional[bool] = None
    # ---------------------------------- 实验 ----------------------------------
    # 实验模式
    resume: Literal['must', 'allow', 'never'] = 'never'
    # 实验名称
    run_name: Optional[str] = None
    # 实验颜色
    run_colors: Optional[List[str]] = None
    # 实验标签
    tags: Optional[List[str]] = None
    # 实验描述
    description: Optional[str] = None
    # 实验运行 ID
    run_id: Optional[str] = None
    # 当前实验是否为新实验
    new: Optional[bool] = None
    # 恢复实验时，云端实验的 config 设置
    config: Optional[dict] = None
    # 恢复实验时，云端实验的指标数据，key -> (column_type, column_class, error, latest step)
    metrics: Optional[RemoteMetric] = None
    # 恢复实验时，云端实验的日志条数
    log_epoch: Optional[int] = None

    # ---------------------------------- 目录 ----------------------------------
    # 是否为临时目录，标识一些运行时环境
    tmp_dir: Optional[bool] = None
    # 日志存放目录
    swanlog_dir: Optional[str] = None
    # 运行目录
    run_dir: Optional[str] = None

    @property
    def backup_file(self):
        assert os.path.exists(self.run_dir), "Run directory does not exist when accessing backup file."
        return os.path.join(self.run_dir, "backup.swanlab")

    @property
    def log_dir(self):
        assert os.path.exists(self.run_dir), "Run directory does not exist when accessing log directory."
        return os.path.join(self.run_dir, "logs")

    @property
    def console_dir(self):
        assert os.path.exists(self.run_dir), "Run directory does not exist when accessing console directory."
        return os.path.join(self.run_dir, "console")

    @property
    def file_dir(self):
        assert os.path.exists(self.run_dir), "Run directory does not exist when accessing file directory."
        return os.path.join(self.run_dir, "files")

    @property
    def media_dir(self):
        assert os.path.exists(self.run_dir), "Run directory does not exist when accessing media directory."
        return os.path.join(self.run_dir, "media")


run_store = RunStore()


def inside(func):
    """
    检查当前代码是否在 swanlab/data 模块下运行
    如果 swanlab 正在运行测试，则允许在测试中调用此函数
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        frame = inspect.currentframe()
        try:
            caller_module = frame.f_back.f_globals.get('__name__', '')
            if not caller_module.startswith('swanlab.data'):
                if 'PYTEST_VERSION' not in os.environ:
                    raise RuntimeError("This function can only be called from swanlab.data module.")
        finally:
            del frame
        return func(*args, **kwargs)

    return wrapper


@inside
def get_run_store():
    """
    此模块只允许在 swanlab/data 模块下访问
    为了提高性能，建议尽量减少对此函数的调用次数
    """
    global run_store
    return run_store


@inside
def reset_run_store():
    global run_store
    run_store = RunStore()
    return None


__all__ = ["get_run_store", "reset_run_store", "RunStore", "RemoteMetric"]



"""
// @author: ComPleHN
// @file: pie.vue
// @time: 2025/5/28 13:12
// @description: 本文件是对于echarts的 饼状图 图表的测试
"""
# ---------------------------------------------- Pie - Pie_set_color ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker

c1 = (
    Pie()
    .add("", [list(z) for z in zip(Faker.choose(), Faker.values())])
    .set_colors(["blue", "green", "yellow", "red", "pink", "orange", "purple"])
    .set_global_opts(title_opts=opts.TitleOpts(title="Pie-设置颜色"))
    .set_series_opts(label_opts=opts.LabelOpts(formatter="{b}: {c}"))
)

# ---------------------------------------------- Pie - Customized_pie ----------------------------------------------
import pyecharts.options as opts
from pyecharts.charts import Pie

"""
Gallery 使用 pyecharts 1.1.0
参考地址: https://echarts.apache.org/examples/editor.html?c=pie-doughnut

目前无法实现的功能:

1、迷之颜色映射的问题
"""

x_data = ["直接访问", "邮件营销", "联盟广告", "视频广告", "搜索引擎"]
y_data = [335, 310, 274, 235, 400]
data_pair = [list(z) for z in zip(x_data, y_data)]
data_pair.sort(key=lambda x: x[1])

c2 = (
    Pie(init_opts=opts.InitOpts(bg_color="#2c343c"))
    .add(
        series_name="访问来源",
        data_pair=data_pair,
        rosetype="radius",
        radius="55%",
        center=["50%", "50%"],
        label_opts=opts.LabelOpts(is_show=False, position="center"),
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(
            title="Customized Pie",
            pos_left="center",
            pos_top="20",
            title_textstyle_opts=opts.TextStyleOpts(color="#fff"),
        ),
        legend_opts=opts.LegendOpts(is_show=False),
    )
    .set_series_opts(
        tooltip_opts=opts.TooltipOpts(
            trigger="item", formatter="{a} <br/>{b}: {c} ({d}%)"
        ),
        label_opts=opts.LabelOpts(color="rgba(255, 255, 255, 0.3)"),
    )
)

# ---------------------------------------------- Pie - Pie_rich_label ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker

c3 = (
    Pie()
    .add(
        "",
        [list(z) for z in zip(Faker.choose(), Faker.values())],
        radius=["40%", "55%"],
        label_opts=opts.LabelOpts(
            position="outside",
            formatter="{a|{a}}{abg|}\n{hr|}\n {b|{b}: }{c}  {per|{d}%}  ",
            background_color="#eee",
            border_color="#aaa",
            border_width=1,
            border_radius=4,
            rich={
                "a": {"color": "#999", "lineHeight": 22, "align": "center"},
                "abg": {
                    "backgroundColor": "#e3e3e3",
                    "width": "100%",
                    "align": "right",
                    "height": 22,
                    "borderRadius": [4, 4, 0, 0],
                },
                "hr": {
                    "borderColor": "#aaa",
                    "width": "100%",
                    "borderWidth": 0.5,
                    "height": 0,
                },
                "b": {"fontSize": 16, "lineHeight": 33},
                "per": {
                    "color": "#eee",
                    "backgroundColor": "#334455",
                    "padding": [2, 4],
                    "borderRadius": 2,
                },
            },
        ),
    )
    .set_global_opts(title_opts=opts.TitleOpts(title="Pie-富文本示例"))
)

# ---------------------------------------------- Pie - Pie_position ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker

c4 = (
    Pie()
    .add(
        "",
        [list(z) for z in zip(Faker.choose(), Faker.values())],
        center=["35%", "50%"],
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(title="Pie-调整位置"),
        legend_opts=opts.LegendOpts(pos_left="15%"),
    )
    .set_series_opts(label_opts=opts.LabelOpts(formatter="{b}: {c}"))
)

# ---------------------------------------------- Pie - Pie_base ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker

c5 = (
    Pie()
    .add("", [list(z) for z in zip(Faker.choose(), Faker.values())])
    .set_global_opts(title_opts=opts.TitleOpts(title="Pie-基本示例"))
    .set_series_opts(label_opts=opts.LabelOpts(formatter="{b}: {c}"))
)

# ---------------------------------------------- Pie - Pie_scroll_legend ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker

c6 = (
    Pie()
    .add(
        "",
        [
            list(z)
            for z in zip(
                Faker.choose() + Faker.choose() + Faker.choose(),
                Faker.values() + Faker.values() + Faker.values(),
            )
        ],
        center=["40%", "50%"],
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(title="Pie-Legend 滚动"),
        legend_opts=opts.LegendOpts(type_="scroll", pos_left="80%", orient="vertical"),
    )
    .set_series_opts(label_opts=opts.LabelOpts(formatter="{b}: {c}"))
)

# ---------------------------------------------- Pie - Doughnut_chart ----------------------------------------------
import pyecharts.options as opts
from pyecharts.charts import Pie

"""
Gallery 使用 pyecharts 1.1.0
参考地址: https://echarts.apache.org/examples/editor.html?c=pie-doughnut

目前无法实现的功能:

1、饼状图中间的图例名称暂时无法显示
"""

x_data = ["直接访问", "邮件营销", "联盟广告", "视频广告", "搜索引擎"]
y_data = [335, 310, 234, 135, 1548]

c7 = (
    Pie()
    .add(
        series_name="访问来源",
        data_pair=[list(z) for z in zip(x_data, y_data)],
        radius=["50%", "70%"],
        label_opts=opts.LabelOpts(is_show=False, position="center"),
    )
    .set_global_opts(legend_opts=opts.LegendOpts(pos_left="legft", orient="vertical"))
    .set_series_opts(
        tooltip_opts=opts.TooltipOpts(
            trigger="item", formatter="{a} <br/>{b}: {c} ({d}%)"
        ),
        # label_opts=opts.LabelOpts(formatter="{b}: {c}")
    )
)

# ---------------------------------------------- Pie - Nested_pies ----------------------------------------------
import pyecharts.options as opts
from pyecharts.charts import Pie

"""
Gallery 使用 pyecharts 1.1.0
参考地址: https://echarts.apache.org/examples/editor.html?c=pie-nest

目前无法实现的功能:

1、暂无
"""

inner_x_data = ["直达", "营销广告", "搜索引擎"]
inner_y_data = [335, 679, 1548]
inner_data_pair = [list(z) for z in zip(inner_x_data, inner_y_data)]

outer_x_data = ["直达", "营销广告", "搜索引擎", "邮件营销", "联盟广告", "视频广告", "百度", "谷歌", "必应", "其他"]
outer_y_data = [335, 310, 234, 135, 1048, 251, 147, 102]
outer_data_pair = [list(z) for z in zip(outer_x_data, outer_y_data)]

c8 = (
    Pie()
    .add(
        series_name="访问来源",
        data_pair=inner_data_pair,
        radius=[0, "30%"],
        label_opts=opts.LabelOpts(position="inner"),
    )
    .add(
        series_name="访问来源",
        radius=["40%", "55%"],
        data_pair=outer_data_pair,
        label_opts=opts.LabelOpts(
            position="outside",
            formatter="{a|{a}}{abg|}\n{hr|}\n {b|{b}: }{c}  {per|{d}%}  ",
            background_color="#eee",
            border_color="#aaa",
            border_width=1,
            border_radius=4,
            rich={
                "a": {"color": "#999", "lineHeight": 22, "align": "center"},
                "abg": {
                    "backgroundColor": "#e3e3e3",
                    "width": "100%",
                    "align": "right",
                    "height": 22,
                    "borderRadius": [4, 4, 0, 0],
                },
                "hr": {
                    "borderColor": "#aaa",
                    "width": "100%",
                    "borderWidth": 0.5,
                    "height": 0,
                },
                "b": {"fontSize": 16, "lineHeight": 33},
                "per": {
                    "color": "#eee",
                    "backgroundColor": "#334455",
                    "padding": [2, 4],
                    "borderRadius": 2,
                },
            },
        ),
    )
    .set_global_opts(legend_opts=opts.LegendOpts(pos_left="left", orient="vertical"))
    .set_series_opts(
        tooltip_opts=opts.TooltipOpts(
            trigger="item", formatter="{a} <br/>{b}: {c} ({d}%)"
        )
    )
)

# ---------------------------------------------- Pie - Mutiple_pie ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.commons.utils import JsCode


fn = """
    function(params) {
        if(params.name == '其他')
            return '\\n\\n\\n' + params.name + ' : ' + params.value + '%';
        return params.name + ' : ' + params.value + '%';
    }
    """


def new_label_opts():
    return opts.LabelOpts(formatter=JsCode(fn), position="center")


c9 = (
    Pie()
    .add(
        "",
        [list(z) for z in zip(["剧情", "其他"], [25, 75])],
        center=["20%", "30%"],
        radius=[60, 80],
        label_opts=new_label_opts(),
    )
    .add(
        "",
        [list(z) for z in zip(["奇幻", "其他"], [24, 76])],
        center=["55%", "30%"],
        radius=[60, 80],
        label_opts=new_label_opts(),
    )
    .add(
        "",
        [list(z) for z in zip(["爱情", "其他"], [14, 86])],
        center=["20%", "70%"],
        radius=[60, 80],
        label_opts=new_label_opts(),
    )
    .add(
        "",
        [list(z) for z in zip(["惊悚", "其他"], [11, 89])],
        center=["55%", "70%"],
        radius=[60, 80],
        label_opts=new_label_opts(),
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(title="Pie-多饼图基本示例"),
        legend_opts=opts.LegendOpts(
            type_="scroll", pos_top="20%", pos_left="80%", orient="vertical"
        ),
    )
)

# ---------------------------------------------- Pie - Pie_rosetype ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker


v = Faker.choose()
c10 = (
    Pie()
    .add(
        "",
        [list(z) for z in zip(v, Faker.values())],
        radius=["30%", "75%"],
        center=["25%", "50%"],
        rosetype="radius",
        label_opts=opts.LabelOpts(is_show=False),
    )
    .add(
        "",
        [list(z) for z in zip(v, Faker.values())],
        radius=["30%", "75%"],
        center=["75%", "50%"],
        rosetype="area",
    )
    .set_global_opts(title_opts=opts.TitleOpts(title="Pie-玫瑰图示例"))
)

# ---------------------------------------------- Pie - Pie_radius ----------------------------------------------
from pyecharts import options as opts
from pyecharts.charts import Pie
from pyecharts.faker import Faker

c11 = (
    Pie()
    .add(
        "",
        [list(z) for z in zip(Faker.choose(), Faker.values())],
        radius=["40%", "75%"],
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(title="Pie-Radius"),
        legend_opts=opts.LegendOpts(orient="vertical", pos_top="15%", pos_left="2%"),
    )
    .set_series_opts(label_opts=opts.LabelOpts(formatter="{b}: {c}"))
)

import swanlab

swanlab.init(
    project="echarts-test",
    experiment_name="pie",
    public=True,
)

swanlab.log(
    {
        "Pie - Pie_set_color": c1,
        "Pie - Customized_pie": c2,
        "Pie - Pie_rich_label": c3,
        "Pie - Pie_position": c4,
        "Pie - Pie_base": c5,
        "Pie - Pie_scroll_legend": c6,
        "Pie - Doughnut_chart": c7,
        "Pie - Nested_pies": c8,
        "Pie - Mutiple_pie": c9,
        "Pie - Pie_rosetype": c10,
        "Pie - Pie_radius": c11
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/3 15:42
@File: pytest_wrapper.py
@IDE: pycharm
@Description:
    测试包装器
"""
import swanlab.data.modules as M


class TestWrapperLine:
    def test_ok(self):
        """
        正常情况
        """
        data = M.Line(1)
        wrapper = M.DataWrapper("test", [data])
        assert wrapper.is_line is True



import os.path
from sys import stdout

from rich.status import Status

from .mlflow import sync_mlflow
from .tensorboard import sync_tensorboardX, sync_tensorboard_torch
from .wandb import sync_wandb

__all__ = ["sync_wandb", "sync_tensorboardX", "sync_tensorboard_torch", "sync_mlflow", "sync"]

from ..core_python import get_client
from ..data.namer import generate_colors
from ..data.porter import DataPorter
from ..log import swanlog


def sync(
    dir_path: str,
    workspace: str = None,
    project_name: str = None,
    raise_error: bool = True,
    login_required: bool = True,
):
    """
    Syncs backup files to the cloud. Before syncing, you must log in.
    :param dir_path: The directory path to sync.
    :param workspace: The workspace to sync the logs to. If not specified, it will use the default workspace.
    :param project_name: The project to sync the logs to. If not specified, it will use the default project.
    :param raise_error: Whether to raise an error if error occurs when syncing.
    :param login_required: Whether login is required before syncing, just for debugging.
    """
    # 第一部分，处理备份文件，读取备份信息到内存中
    try:
        assert os.path.exists(dir_path), f"Directory {dir_path} does not exist."
        try:
            client = get_client()
        except ValueError:
            client = None
            assert not login_required, "Please log in first, use `swanlab login` to log in."
        stdout.flush()
        with Status("🔁 Syncing...", spinner="dots"):
            with DataPorter().open_for_sync(run_dir=dir_path) as porter:
                project, experiment = porter.parse()
                assert client is not None, "Please log in first, use `swanlab login` to log in."
                client.mount_project(
                    name=project_name or project.name,
                    username=workspace or project.workspace,
                    public=project.public,
                )
                colors = generate_colors(client.history_exp_count)
                client.mount_exp(
                    exp_name=experiment.name,
                    colors=colors,
                    description=experiment.description,
                    tags=experiment.tags,
                )
                success = porter.synchronize()
                # 3.5 更新实验状态
                client.update_state(success=success)
        swanlog.info("🚀 Sync completed, View run at ", client.web_exp_url)
    except Exception as e:
        if raise_error:
            raise e
        else:
            return swanlog.error(f"😢 Error syncing: {e}")



"""
// @author: ComPleHN
// @file: calendar.vue
// @time: 2025/5/27 13:29
// @description: 本文件是对于echarts的 Calendar 图表测试 , 文件名不叫calendar是因为和库文件重名
"""
# ---------------------------------------------- Calendar - Calendar_heatmap ----------------------------------------------
import random
import datetime

import pyecharts.options as opts
from pyecharts.charts import Calendar


begin = datetime.date(2017, 1, 1)
end = datetime.date(2017, 12, 31)
data = [
    [str(begin + datetime.timedelta(days=i)), random.randint(1000, 25000)]
    for i in range((end - begin).days + 1)
]

c1 = (
    Calendar()
    .add(
        series_name="",
        yaxis_data=data,
        calendar_opts=opts.CalendarOpts(
            pos_top="120",
            pos_left="30",
            pos_right="30",
            range_="2017",
            yearlabel_opts=opts.CalendarYearLabelOpts(is_show=False),
        ),
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(pos_top="30", pos_left="center", title="2017年步数情况"),
        visualmap_opts=opts.VisualMapOpts(
            max_=20000, min_=500, orient="horizontal", is_piecewise=False
        ),
    )
)

# ---------------------------------------------- Calendar - Calendar_label_setting ----------------------------------------------
import datetime
import random

from pyecharts import options as opts
from pyecharts.charts import Calendar


begin = datetime.date(2017, 1, 1)
end = datetime.date(2017, 12, 31)
data = [
    [str(begin + datetime.timedelta(days=i)), random.randint(1000, 25000)]
    for i in range((end - begin).days + 1)
]

c2 = (
    Calendar()
    .add(
        "",
        data,
        calendar_opts=opts.CalendarOpts(
            range_="2017",
            daylabel_opts=opts.CalendarDayLabelOpts(name_map="cn"),
            monthlabel_opts=opts.CalendarMonthLabelOpts(name_map="cn"),
        ),
    )
    .set_global_opts(
        title_opts=opts.TitleOpts(title="Calendar-2017年微信步数情况(中文 Label)"),
        visualmap_opts=opts.VisualMapOpts(
            max_=20000,
            min_=500,
            orient="horizontal",
            is_piecewise=True,
            pos_top="230px",
            pos_left="100px",
        ),
    )
)

# ---------------------------------------------- Calendar - Calendar_base ----------------------------------------------
import datetime
import random

from pyecharts import options as opts
from pyecharts.charts import Calendar


begin = datetime.date(2017, 1, 1)
end = datetime.date(2017, 12, 31)
data = [
    [str(begin + datetime.timedelta(days=i)), random.randint(1000, 25000)]
    for i in range((end - begin).days + 1)
]

c3 = (
    Calendar()
    .add("", data, calendar_opts=opts.CalendarOpts(range_="2017"))
    .set_global_opts(
        title_opts=opts.TitleOpts(title="Calendar-2017年微信步数情况"),
        visualmap_opts=opts.VisualMapOpts(
            max_=20000,
            min_=500,
            orient="horizontal",
            is_piecewise=True,
            pos_top="230px",
            pos_left="100px",
        ),
    )
)



import swanlab

swanlab.init(
    project="echarts-test",
    experiment_name="calendar",
    public=True,
)

swanlab.log(
    {
        "Calendar - Calendar_heatmap": c1,
        "Calendar - Calendar_label_setting": c2,
        "Calendar - Calendar_base": c3
    }
)






import numpy as np
import numbers
from typing import Mapping, Any


class BoundingBoxes:
    """
    用于处理2D边界框（bounding boxes）数据的类。
    """

    # box_item的格式形如：{"box_data":..., "class_labels":...}
    # box_data存储了所有box的数据(包括位置、类别id、准确率、是否有caption、...)，class_labels存储了所有的标签合集

    def __init__(self, box_item: dict, key: str) -> None:
        # 验证box_item是否符合要求
        self.validate(box_item)

        self._box_data = box_item["box_data"]
        self._key = key

        # 如果box_item中不存在class_labels参数, 则生成默认格式的类型标签
        if "class_labels" not in box_item:
            classes = np.unique(list(box["class_id"] for box in box_item["box_data"])).astype(np.int32).tolist()
            class_labels = {c: "class_" + str(c) for c in classes}
            self._class_labels = class_labels
        else:
            self._class_labels = box_item["class_labels"]

    def validate(self, box_item) -> bool:
        """
        检查box_data是否符合要求
        """

        # 如果box_item中包含class_labels参数，则检查参数是否符合规范
        if "class_labels" in box_item:
            for k, v in list(box_item["class_labels"].items()):
                if (not isinstance(k, numbers.Number)) or (not isinstance(v, str)):
                    raise TypeError("Class labels must be a dictionary of numbers to string")

        boxes = box_item["box_data"]

        # 如果boxes不是list类型，则报错
        if not isinstance(boxes, list):
            raise TypeError("'box_data' must be a list")

        # 对于boxes中的每一个box
        for box in boxes:
            error_str = "Each box in box_data must contain a position with: middle, width, and height or \
                    \nminX, maxX, minY, maxY."

            # 验证position参数（必须参数）
            # 如果box中不包含position参数，则报错
            if "position" not in box:
                raise TypeError(error_str)
            else:
                valid = False
                if (
                    "middle" in box["position"]
                    and len(box["position"]["middle"]) == 2
                    and validate_dict_key_has_number(box["position"], "width")
                    and validate_dict_key_has_number(box["position"], "height")
                ):
                    valid = True

                elif (
                    validate_dict_key_has_number(box["position"], "minX")
                    and validate_dict_key_has_number(box["position"], "maxX")
                    and validate_dict_key_has_number(box["position"], "minY")
                    and validate_dict_key_has_number(box["position"], "maxY")
                ):
                    valid = True

                if not valid:
                    raise TypeError(error_str)

            # 验证score参数（可选参数）
            if ("scores" in box) and not isinstance(box["scores"], dict):
                raise TypeError("Box scores must be a dictionary")
            elif "scores" in box:
                for k, v in list(box["scores"].items()):
                    if not isinstance(k, str):
                        raise TypeError("A score key must be a string")
                    if not isinstance(v, numbers.Number):
                        raise TypeError("A score value must be a number")

            # 验证class_id参数
            if ("class_id" in box) and not isinstance(box["class_id"], int):
                raise TypeError("A box's class_id must be an integer")

            # 验证box_caption参数
            if ("box_caption" in box) and not isinstance(box["box_caption"], str):
                raise TypeError("A box's caption must be a string")

        return True


def validate_dict_key_has_number(dictionary: Mapping, key: Any) -> bool:
    return key in dictionary and isinstance(dictionary[key], numbers.Number)



"""
@author: cunyue
@file: datastore.py
@time: 2025/6/5 16:32
@description: 记录的数据遵循LevelDB格式：https://github.com/google/leveldb/blob/main/doc/log_format.md
我们使用 crc32 计算数据校验和，crc32 相对轻量，且计算速度较快
字符编码使用 utf-8, 确保数据兼容性 （这与 LevelDB 的规范有所冲突，如果有必要，未来可以升级版本并通过LEVELDBLOG_HEADER_VERSION兼容）
这为后续引入 protobuf 或其他序列化格式打下基础
DataStore 大致代码借鉴自 W&B

文件头版本区分了不同更新时的格式变化，这方面我们不会做向下兼容，即低版本文件不一定能被新版本读取，可以通过版本降级来区分不同版本号
"""

import os
import struct
import zlib
from typing import Optional, Any, IO, Tuple

LEVELDBLOG_HEADER_LEN = 7
LEVELDBLOG_BLOCK_LEN = 32768
LEVELDBLOG_DATA_LEN = LEVELDBLOG_BLOCK_LEN - LEVELDBLOG_HEADER_LEN

LEVELDBLOG_FULL = 1
LEVELDBLOG_FIRST = 2
LEVELDBLOG_MIDDLE = 3
LEVELDBLOG_LAST = 4


LEVELDBLOG_HEADER_IDENT = ":SWL"
LEVELDBLOG_HEADER_MAGIC = 0xE1D6  # zlib.crc32(bytes("SwanLab", 'utf-8')) & 0xffff
LEVELDBLOG_HEADER_VERSION = 0


def strtobytes(x):
    """
    文件转字符串
    """
    return bytes(x, "utf-8")


def bytestostr(x):
    return str(x, 'utf-8')


class DataStore:

    def __init__(self):
        self._filename: Optional[str] = None
        self._fp: Optional[IO[Any]] = None
        # 当前文件的偏移量
        self._index: int = 0
        # 当前文件的已刷写偏移量
        self._flush_offset = 0
        # 日志系统预计算并缓存CRC32校验值，缓存每一个数据类型的CRC32值，分别存在各自的索引位置
        self._crc = [0] * (LEVELDBLOG_LAST + 1)
        for x in range(1, LEVELDBLOG_LAST + 1):
            self._crc[x] = zlib.crc32(strtobytes(chr(x))) & 0xFFFFFFFF

        # 是否为扫描模式打开文件
        self._opened_for_scan = False
        # 当前文件大小（仅在扫描模式下有效）
        self._size_bytes: int = 0

    # ---------------------------------- 读取 ----------------------------------

    def open_for_scan(self, filename: str):
        self._filename = filename
        self._fp = open(filename, "r+b")
        self._index = 0
        self._size_bytes = os.stat(filename).st_size
        self._opened_for_scan = True
        self._read_header()

    def _read_header(self):
        header = self._fp.read(LEVELDBLOG_HEADER_LEN)
        assert (
            len(header) == LEVELDBLOG_HEADER_LEN
        ), f"header is {len(header)} bytes instead of the expected {LEVELDBLOG_HEADER_LEN}"
        ident, magic, version = struct.unpack("<4sHB", header)
        if ident != strtobytes(LEVELDBLOG_HEADER_IDENT):
            raise Exception("Invalid header")
        if magic != LEVELDBLOG_HEADER_MAGIC:
            raise Exception("Invalid header")
        if version != LEVELDBLOG_HEADER_VERSION:
            raise Exception(f"Invalid backup version: {version}, please check your swanlab version.")
        self._index += len(header)

    def _scan_record(self) -> Optional[Tuple[int, bytes]]:
        """
        扫描一条记录
        """
        assert self._opened_for_scan, "file not open for scanning"
        # 1. 读取数据头
        header = self._fp.read(LEVELDBLOG_HEADER_LEN)
        if len(header) == 0:
            return None
        assert (
            len(header) == LEVELDBLOG_HEADER_LEN
        ), f"record header is {len(header)} bytes instead of the expected {LEVELDBLOG_HEADER_LEN}"
        # 2. 解析数据头并校验数据完整性
        checksum, data_length, data_type = struct.unpack("<IHB", header)
        self._index += LEVELDBLOG_HEADER_LEN
        data = self._fp.read(data_length)
        checksum_computed = zlib.crc32(data, self._crc[data_type]) & 0xFFFFFFFF
        assert checksum == checksum_computed, "record checksum is invalid, data may be corrupt"
        self._index += data_length
        # 3. 返回数据
        return int(data_type), data

    def scan(self) -> Optional[str]:
        """
        扫描日志文件，返回一条记录
        """
        # 1. 一次读取一条记录，如果剩余空间不足存储数据头，校验并跳过，此为写入的逆操作
        offset = self._index % LEVELDBLOG_BLOCK_LEN
        space_left = LEVELDBLOG_BLOCK_LEN - offset
        if space_left < LEVELDBLOG_HEADER_LEN:
            pad_check = strtobytes("\x00" * space_left)
            pad = self._fp.read(space_left)
            # 校验必须为0
            assert pad == pad_check, "invalid padding"
            self._index += space_left
        # 2. 扫描一条记录
        record = self._scan_record()
        if record is None:  # eof
            return None
        dtype, data = record
        if dtype == LEVELDBLOG_FULL:
            return bytestostr(data)
        # 3. 如果是第一条记录，则继续扫描直到找到最后一条记录
        assert dtype == LEVELDBLOG_FIRST, f"expected record to be type {LEVELDBLOG_FIRST} but found {dtype}"
        while True:
            record = self._scan_record()
            if record is None:  # eof
                return None
            dtype, new_data = record
            if dtype == LEVELDBLOG_LAST:
                data += new_data
                break
            assert dtype == LEVELDBLOG_MIDDLE, f"expected record to be type {LEVELDBLOG_MIDDLE} but found {dtype}"
            data += new_data
        return bytestostr(data)

    def __iter__(self):
        """
        实现迭代器接口，允许使用 for 循环遍历日志文件，仅在文件已打开并且处于扫描模式时有效
        """
        assert self._opened_for_scan, "file not open for scanning, cannot iterate"
        return self

    def __next__(self):
        record = self.scan()
        if record is None:
            raise StopIteration("End of file reached")
        return record

    # ---------------------------------- 写入 ----------------------------------

    def open_for_write(self, filename: str):
        self._filename = filename
        self._fp = open(filename, "xb")
        # 写入文件头, 长度等于 LEVELDBLOG_HEADER_LEN
        data = struct.pack(
            "<4sHB",
            strtobytes(LEVELDBLOG_HEADER_IDENT),
            LEVELDBLOG_HEADER_MAGIC,
            LEVELDBLOG_HEADER_VERSION,
        )
        assert len(data) == LEVELDBLOG_HEADER_LEN, f"header size is {len(data)} bytes, expected {LEVELDBLOG_HEADER_LEN}"
        self._fp.write(data)
        self._index += len(data)

    def _write_record(self, data: bytes, data_type: int = LEVELDBLOG_FULL):
        """
        写入记录到日志文件
        """
        assert len(data) + LEVELDBLOG_HEADER_LEN <= (
            LEVELDBLOG_BLOCK_LEN - self._index % LEVELDBLOG_BLOCK_LEN
        ), "not enough space to write new records"
        data_length = len(data)
        # 计算校验值，校验值为对数据和数据类型的 CRC32 校验和
        checksum = zlib.crc32(data, self._crc[data_type]) & 0xFFFFFFFF
        # 写入数据头，格式为：<IHB>，分别表示校验和、数据长度和数据类型
        # I: unsigned int (4 bytes), H: unsigned short (2 bytes), B: unsigned char (1 byte)
        self._fp.write(struct.pack("<IHB", checksum, data_length, data_type))
        if data_length:
            self._fp.write(data)
        self._index += LEVELDBLOG_HEADER_LEN + len(data)

    def write(self, s: str):
        """
        写入数据到日志文件，遵循 LevelDB 规范
        :param s: 要写入的数据，必须是字符串形式
        :return: 返回写入的起始偏移量、当前偏移量和已刷写偏移量
        """
        data = strtobytes(s)
        # 1. 计算偏移量
        start_offset = self._index
        offset = self._index % LEVELDBLOG_BLOCK_LEN
        space_left = LEVELDBLOG_BLOCK_LEN - offset
        data_used = 0
        data_left = len(data)
        # 2. 剩余长度小于数据头长度则填充0，归位到下一个块
        if space_left < LEVELDBLOG_HEADER_LEN:
            pad = "\x00" * space_left
            self._fp.write(strtobytes(pad))
            self._index += space_left
            space_left = LEVELDBLOG_BLOCK_LEN
        # 3. 如果剩余长度大于等于数据长度，则直接写入
        if data_left + LEVELDBLOG_HEADER_LEN <= space_left:
            self._write_record(data)
        # 4. 否则需要分块写入（注意此时我们可能在一个块的中间）
        else:
            # 4.1 写入第一个数据块，确保接下来数据独占一个块
            data_room = space_left - LEVELDBLOG_HEADER_LEN
            self._write_record(data[:data_room], LEVELDBLOG_FIRST)
            data_used += data_room
            data_left -= data_room
            assert data_left, "data_left should be non-zero"
            # 4.2 写入中间数据
            while data_left > LEVELDBLOG_DATA_LEN:
                self._write_record(
                    data[data_used : data_used + LEVELDBLOG_DATA_LEN],
                    LEVELDBLOG_MIDDLE,
                )
                data_used += LEVELDBLOG_DATA_LEN
                data_left -= LEVELDBLOG_DATA_LEN
            # 4.3 写入最后一个数据块
            self._write_record(data[data_used:], LEVELDBLOG_LAST)
            # 刷写完整数据
            try:
                self._fp.flush()
                os.fsync(self._fp.fileno())
            except OSError:
                # 如果操作系统不支持 fsync，可能会抛出 OSError，忽略此错误即可
                pass
            self._flush_offset = self._index

        return start_offset, self._index, self._flush_offset

    # ---------------------------------- 辅助函数 ----------------------------------

    def ensure_flushed(self) -> None:
        self._fp.flush()

    def close(self):
        # 关闭文件句柄
        self._fp.close()



import swanlab

def _extract_args(args, kwargs, param_names):
    """
    从args和kwargs中提取参数值的通用函数
    
    Args:
        args: 位置参数元组
        kwargs: 关键字参数字典
        param_names: 参数名称列表
    
    Returns:
        tuple: 按param_names顺序返回提取的参数值
    """
    values = []
    for i, name in enumerate(param_names):
        if len(args) > i:
            values.append(args[i])
        else:
            values.append(kwargs.get(name, None))
    return tuple(values)


def _create_patched_methods(SummaryWriter, logdir_extractor):
    """
    创建patched方法的工厂函数
    
    Args:
        SummaryWriter: SummaryWriter类
        logdir_extractor: 提取logdir的函数
    
    Returns:
        tuple: (patched_init, patched_add_scalar, patched_add_image, patched_close)
    """
    original_init = SummaryWriter.__init__
    original_add_scalar = SummaryWriter.add_scalar
    original_add_scalars = SummaryWriter.add_scalars
    original_add_image = SummaryWriter.add_image
    original_add_text = SummaryWriter.add_text
    original_close = SummaryWriter.close

    def patched_init(self, *args, **kwargs):
        tb_logdir = logdir_extractor(args, kwargs)
        
        tb_config = {
            'tensorboard_logdir': tb_logdir,
        }

        if swanlab.data.get_run() is None:
            swanlab.init(config=tb_config)
        else:
            swanlab.config.update(tb_config)

        return original_init(self, *args, **kwargs)

    def patched_add_scalar(self, *args, **kwargs):
        tag, scalar_value, global_step = _extract_args(
            args, kwargs, ['tag', 'scalar_value', 'global_step']
        )
        
        data = {tag: scalar_value}
        swanlab.log(data=data, step=global_step)
        
        return original_add_scalar(self, *args, **kwargs)

    def patched_add_scalars(self, *args, **kwargs):
        # writer.add_scalars('Loss', {'train': loss_train, 'val': loss_val}, global_step=step)
        tag, scalar_value_dict, global_step = _extract_args(
            args, kwargs, ['tag', 'scalar_value_dict', 'global_step']
        )
        for dict_tag, value in scalar_value_dict.items():
            data = {f"{tag}/{dict_tag}": value}
            swanlab.log(data=data, step=global_step)
        return original_add_scalars(self, *args, **kwargs)

    def patched_add_image(self, *args, **kwargs):
        import numpy as np
        
        tag, img_tensor, global_step, dataformats = _extract_args(
            args, kwargs, ['tag', 'img_tensor', 'global_step', 'dataformats']
        )
        dataformats = dataformats or 'CHW'  # 设置默认值
        
        # Convert to numpy array if it's a tensor
        if hasattr(img_tensor, 'cpu'):
            img_tensor = img_tensor.cpu()
        if hasattr(img_tensor, 'numpy'):
            img_tensor = img_tensor.numpy()
            
        # Handle different input formats
        if dataformats == 'CHW':
            # Convert CHW to HWC for swanlab
            img_tensor = np.transpose(img_tensor, (1, 2, 0))
        elif dataformats == 'NCHW':
            # Take first image if batch dimension exists and convert to HWC
            img_tensor = np.transpose(img_tensor, (1, 2, 0))
        elif dataformats == 'HW':
            # Add channel dimension for grayscale
            img_tensor = np.expand_dims(img_tensor, axis=-1)
        elif dataformats == 'HWC':
            # Already in correct format
            pass
            
        data = {tag: swanlab.Image(img_tensor)}
        swanlab.log(data=data, step=global_step)

        return original_add_image(self, *args, **kwargs)
    
    def patched_add_text(self, *args, **kwargs):
        tag, text_string, global_step = _extract_args(
            args, kwargs, ['tag', 'text_string', 'global_step']
        )
        data = {tag: swanlab.Text(text_string)}
        swanlab.log(data=data, step=global_step)
        return original_add_text(self, *args, **kwargs)
    

    def patched_close(self):
        # 调用原始的close方法
        original_close(self)
        # 关闭SwanLab记录器
        swanlab.finish()

    return patched_init, patched_add_scalar, patched_add_scalars, patched_add_image, patched_add_text, patched_close


def _apply_patches(SummaryWriter, patched_methods):
    """
    应用monkey patch到SummaryWriter类
    
    Args:
        SummaryWriter: SummaryWriter类
        patched_methods: (patched_init, patched_add_scalar, patched_add_image, patched_close)
    """
    patched_init, patched_add_scalar, patched_add_scalars, patched_add_image, patched_add_text, patched_close = patched_methods
    
    SummaryWriter.__init__ = patched_init
    SummaryWriter.add_scalar = patched_add_scalar
    SummaryWriter.add_scalars = patched_add_scalars
    SummaryWriter.add_image = patched_add_image
    SummaryWriter.add_text = patched_add_text
    SummaryWriter.close = patched_close


def _sync_tensorboard_generic(import_func, logdir_extractor):
    """
    通用的tensorboard同步函数
    
    Args:
        import_func: 导入SummaryWriter的函数
        logdir_extractor: 提取logdir的函数
    """
    try:
        SummaryWriter = import_func()
    except ImportError as e:
        raise ImportError(f"Import failed: {e}")

    patched_methods = _create_patched_methods(SummaryWriter, logdir_extractor)
    _apply_patches(SummaryWriter, patched_methods)


def sync_tensorboardX():
    """
    同步tensorboardX到swanlab

    from tensorboardX import SummaryWriter
    import numpy as np
    import swanlab

    swanlab.sync_tensorboardX()
    writer = SummaryWriter('runs/example')

    for i in range(100):
        scalar_value = np.random.rand()
        writer.add_scalar('random_scalar', scalar_value, i)

    writer.close()
    """
    def import_tensorboardx():
        from tensorboardX import SummaryWriter
        return SummaryWriter
    
    def extract_logdir_tensorboardx(args, kwargs):
        logdir, _, _, _, _, _, _, log_dir, _ = _extract_args(
            args, kwargs, 
            ['logdir', 'comment', 'purge_step', 'max_queue', 'flush_secs', 
             'filename_suffix', 'write_to_disk', 'log_dir', 'comet_config']
        )
        return logdir or log_dir
    
    _sync_tensorboard_generic(import_tensorboardx, extract_logdir_tensorboardx)


def sync_tensorboard_torch():
    """
    同步torch自带的tensorboard到swanlab

    from torch.utils.tensorboard import SummaryWriter
    import numpy as np
    import swanlab

    swanlab.sync_tensorboard_torch()
    writer = SummaryWriter('runs/example')

    for i in range(100):
        scalar_value = np.random.rand()
        writer.add_scalar('random_scalar', scalar_value, i)

    writer.close()
    """
    def import_torch_tensorboard():
        from torch.utils.tensorboard import SummaryWriter
        return SummaryWriter
    
    def extract_logdir_torch(args, kwargs):
        logdir, _ = _extract_args(args, kwargs, ['log_dir', 'comment'])
        return logdir
    
    _sync_tensorboard_generic(import_torch_tensorboard, extract_logdir_torch)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/26 16:03
@File: test_sdk.py
@IDE: pycharm
@Description:
    测试sdk的一些api
"""
import os
import random

import platformdirs
import pytest
from nanoid import generate

import swanlab.data.sdk as S
import swanlab.data.utils
import swanlab.error as Err
import tutils as T
from swanlab import Settings
from swanlab.core_python import reset_client, get_client
from swanlab.data.run import get_run
from swanlab.data.store import get_run_store
from swanlab.env import SwanLabEnv, get_save_dir
from swanlab.log import swanlog


@pytest.fixture(scope="function", autouse=True)
def setup_function():
    """
    在当前测试文件下的每个测试函数执行前后执行
    """
    swanlog.disable_log()
    run = get_run()
    if run is not None:
        run.finish()
    yield
    run = get_run()
    if run is not None:
        run.finish()
    swanlog.enable_log()


MODE = SwanLabEnv.MODE.value


class TestInitModeFunc:

    def test_init_error_mode(self):
        """
        初始化时mode参数错误
        """
        with pytest.raises(ValueError):
            swanlab.data.utils._init_mode("123456")  # noqa

    @pytest.mark.parametrize("mode", ["disabled", "local", "cloud"])
    def test_init_mode(self, mode, monkeypatch):
        """
        初始化时mode参数正确
        """
        if mode == 'cloud':
            mode = 'local'
            monkeypatch.setattr("builtins.input", lambda _: "3")
        swanlab.data.utils._init_mode(mode)
        assert os.environ[MODE] == mode
        del os.environ[MODE]
        # # 大写
        # S._init_mode(mode.upper())
        # assert os.environ[MODE] == mode

    @pytest.mark.parametrize("mode", ["disabled", "local", "cloud"])
    def test_overwrite_mode(self, mode, monkeypatch):
        """
        初始化时mode参数正确，覆盖环境变量
        """
        if mode == 'cloud':
            mode = 'local'
            monkeypatch.setattr("builtins.input", lambda _: "3")
        os.environ[MODE] = "123456"
        swanlab.data.utils._init_mode(mode)
        assert os.environ[MODE] == mode

    def test_no_api_key_to_cloud(self, monkeypatch):
        """
        初始化时mode为cloud，但是没有设置apikey
        """
        if SwanLabEnv.API_KEY.value in os.environ:
            del os.environ[SwanLabEnv.API_KEY.value]
        monkeypatch.setattr("builtins.input", lambda _: "3")
        mode, login_info = swanlab.data.utils._init_mode("cloud")
        assert mode == "local"
        assert login_info is None

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_init_cloud_with_no_api_key(self, monkeypatch):
        """
        初始化时mode为cloud，但是没有设置apikey
        """
        api_key = os.environ[SwanLabEnv.API_KEY.value]
        del os.environ[SwanLabEnv.API_KEY.value]
        # 在测试时默认会在交互模式下
        # 接下来需要模拟终端连接，使用monkeypatch
        # 三种选择方式：
        # 1. 输入api key
        # 2. 创建账号
        # 3. 使用本地版

        # 选择第三种
        monkeypatch.setattr("builtins.input", lambda _: "3")
        mode, login_info = swanlab.data.utils._init_mode("cloud")
        assert mode == "local"
        assert login_info is None

        # 选择第二种
        monkeypatch.setattr("builtins.input", lambda _: "2")
        monkeypatch.setattr("getpass.getpass", lambda _: api_key)
        mode, login_info = swanlab.data.utils._init_mode("cloud")
        assert mode == "cloud"
        assert login_info is not None

        # 登录后会保存key，测试时需要删除
        os.remove(os.path.join(get_save_dir(), ".netrc"))

        # 选择第一种
        monkeypatch.setattr("builtins.input", lambda _: "1")
        monkeypatch.setattr("getpass.getpass", lambda _: api_key)
        mode, login_info = swanlab.data.utils._init_mode("cloud")
        assert mode == "cloud"
        assert login_info is not None

        # 登录后会保存key，测试时需要删除
        os.remove(os.path.join(get_save_dir(), ".netrc"))

        # 选择其他的
        def create_other_input():
            first = True

            def oi():
                nonlocal first
                if first:
                    first = False
                    return "123456"
                else:
                    return "3"

            return oi

        other_input = create_other_input()
        monkeypatch.setattr("builtins.input", lambda _: other_input())
        mode, login_info = swanlab.data.utils._init_mode("cloud")
        assert mode == "local"
        assert login_info is None


class TestInitMode:
    """
    测试init时函数的mode参数设置行为
    """

    def test_init_local(self):
        run = S.init(mode="local")
        assert os.environ[MODE] == "local"
        run.log({"TestInitMode": 1})  # 不会报错
        assert run.mode == "local"
        assert run.public.cloud.available is False
        assert get_run() is not None
        assert run.public.cloud.project_name is None

    def test_init_disabled(self):
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        run = S.init(mode="disabled", logdir=logdir)
        assert not os.path.exists(logdir)
        assert os.environ[MODE] == "disabled"
        run.log({"TestInitMode": 1})  # 不会报错
        a = run.public.run_dir
        assert os.path.exists(a)
        assert get_run() is not None

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_init_cloud(self):
        S.login(T.is_skip_cloud_test)
        run = S.init(mode="cloud")
        assert os.environ[MODE] == "cloud"
        run.log({"TestInitMode": 1})  # 不会报错
        assert get_run() is not None
        for key in run.public.json()['cloud']:
            assert run.public.json()['cloud'][key] is not None

    def test_init_error(self):
        with pytest.raises(ValueError):
            S.init(mode="123456")  # noqa
        assert get_run() is None

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_init_multiple(self):
        # 先初始化cloud
        self.test_init_cloud()
        get_run().finish()
        # 再初始化local
        self.test_init_local()
        get_run().finish()
        # 再初始化disabled
        self.test_init_disabled()

    # ---------------------------------- 测试环境变量输入 ----------------------------------

    def test_init_disabled_env(self):
        os.environ[MODE] = "disabled"
        run = S.init()
        assert os.environ[MODE] == "disabled"
        run.log({"TestInitMode": 1})
        a = run.public.run_dir
        assert os.path.exists(a)
        assert get_run() is not None

    def test_init_local_env(self):
        os.environ[MODE] = "local"
        run = S.init()
        assert os.environ[MODE] == "local"
        run.log({"TestInitMode": 1})

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_init_cloud_env(self):
        os.environ[MODE] = "cloud"
        S.login(T.is_skip_cloud_test)
        run = S.init()
        assert os.environ[MODE] == "cloud"
        run.log({"TestInitMode": 1})

    # ---------------------------------- 环境变量和mode的情况 ----------------------------------

    def test_init_disabled_env_mode(self):
        os.environ[MODE] = "local"
        run = S.init(mode="disabled")
        assert os.environ[MODE] == "disabled"
        run.log({"TestInitMode": 1})
        a = run.public.run_dir
        assert os.path.exists(a)
        assert get_run() is not None


class TestInitProject:
    """
    测试init时函数的project参数设置行为
    """

    def test_init_project_none(self):
        """
        设置project为None
        """
        run = S.init(project=None, mode="disabled")
        assert run.public.project_name == os.path.basename(os.getcwd())

    def test_init_project(self):
        """
        设置project为字符串
        """
        project = "test_project"
        run = S.init(project=project, mode="disabled")
        assert run.public.project_name == project


def test_reinit():
    """
    测试重新初始化
    """
    run = S.init(mode="disabled")
    run_again = S.init(mode="local")
    assert run == run_again
    reinit = S.init(mode="disabled", reinit=True)
    assert run != reinit
    assert run.running is False
    assert run.success is True
    assert run.crashed is False


LOG_DIR = SwanLabEnv.SWANLOG_FOLDER.value


class TestInitLogdir:
    """
    测试init时函数的logdir参数设置行为
    """

    def test_init_logdir_disabled(self):
        """
        disabled模式下设置logdir不生效，采用的是环境变量的设置
        """
        logdir = generate()
        run = S.init(logdir=logdir, mode="disabled")
        assert run.public.swanlog_dir != logdir
        assert run.public.swanlog_dir == os.environ[LOG_DIR]
        run.finish()
        del os.environ[LOG_DIR]
        run = S.init(logdir=logdir, mode="disabled")
        assert run.public.swanlog_dir != logdir
        assert run.public.swanlog_dir == platformdirs.user_cache_dir(appname="swanlab", appauthor="SwanHubX")
        os.path.exists(run.public.swanlog_dir)

    def test_init_logdir_enabled(self):
        """
        其他模式下设置logdir生效
        """
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        run = S.init(logdir=logdir, mode="local")
        assert run.public.swanlog_dir == logdir
        run.finish()
        del os.environ[LOG_DIR]
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        run = S.init(logdir=logdir, mode="local")
        assert run.public.swanlog_dir == logdir

    def test_init_logdir_env(self):
        """
        通过环境变量设置logdir
        """
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        os.environ[LOG_DIR] = logdir
        run = S.init(mode="local")
        assert run.public.swanlog_dir == logdir
        run.finish()
        del os.environ[LOG_DIR]
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        os.environ[LOG_DIR] = logdir
        run = S.init(mode="local")
        assert run.public.swanlog_dir == logdir

    def test_init_logdir_offline(self):
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        os.environ[LOG_DIR] = logdir
        run = S.init(mode="offline")
        assert run.public.swanlog_dir == logdir
        run.finish()
        del os.environ[LOG_DIR]
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        os.environ[LOG_DIR] = logdir
        run = S.init(mode="offline")
        assert run.public.swanlog_dir == logdir

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_init_logdir_settings_backup(self):
        logdir = os.path.join(T.TEMP_PATH, generate()).__str__()
        os.environ[LOG_DIR] = logdir
        # cloud模式下backup功能默认开启，此时依然生成文件夹
        run = S.init(mode="cloud")
        assert run.public.swanlog_dir == logdir
        os.path.exists(logdir)
        run.finish()
        # 如果通过 settings 设置为 False，则不会生成文件夹
        del os.environ[LOG_DIR]
        run = S.init(mode="cloud", settings=Settings(backup=False))
        assert run.public.swanlog_dir == platformdirs.user_cache_dir(appname="swanlab", appauthor="SwanHubX")
        assert run.public.swanlog_dir != logdir


@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
class TestLogin:
    """
    测试通过sdk封装的login函数登录
    不填apikey的部分不太好测
    """

    @staticmethod
    def get_password(prompt: str):
        # 如果是第一次登录，使用错误的key，会提示重新输入
        if "Paste" in prompt:
            return generate()
        else:
            return T.is_skip_cloud_test

    def test_use_home_key(self, monkeypatch):
        """
        使用家目录下的key，不需要输入
        如果家目录下的key获取失败，会使用getpass.getpass要求用户输入，作为测试，使用monkeypatch替换getpass.getpass
        """
        os.environ[LOG_DIR] = T.TEMP_PATH
        monkeypatch.setattr("getpass.getpass", self.get_password)
        S.login()
        # 默认不保存Key
        assert not os.path.exists(os.path.join(get_save_dir(), ".netrc"))

    def test_use_home_key_save(self, monkeypatch):
        """
        使用家目录下的key，不需要输入
        如果家目录下的key获取失败，会使用getpass.getpass要求用户输入，作为测试，使用monkeypatch替换getpass.getpass
        """
        os.environ[LOG_DIR] = T.TEMP_PATH
        monkeypatch.setattr("getpass.getpass", self.get_password)
        S.login(save=True)
        # 保存Key
        assert os.path.exists(os.path.join(get_save_dir(), ".netrc"))

    def test_use_input_key(self, monkeypatch):
        """
        使用输入的key
        """
        key = generate(size=21, alphabet="12345")
        with pytest.raises(Err.ValidationError):
            S.login(api_key=key)
        key = T.API_KEY
        S.login(api_key=key)

    def test_use_env_key(self, monkeypatch):
        """
        测试code登录，使用环境变量key
        """

        def _():
            raise RuntimeError("this function should not be called")

        monkeypatch.setattr("getpass.getpass", _)
        os.environ[SwanLabEnv.API_KEY.value] = generate(size=21, alphabet="12345")
        with pytest.raises(Err.ValidationError):
            S.login()
        os.environ[SwanLabEnv.API_KEY.value] = T.API_KEY
        S.login()

    def test_pass_host(self):
        """
        传入host参数
        """
        del os.environ[SwanLabEnv.WEB_HOST.value]
        del os.environ[SwanLabEnv.API_HOST.value]
        os.environ[SwanLabEnv.API_KEY.value] = T.API_KEY
        S.login(host=T.API_HOST.rstrip("/api"))
        assert os.environ[SwanLabEnv.API_HOST.value] == T.API_HOST
        assert os.environ.get(SwanLabEnv.WEB_HOST.value) == T.API_HOST.rstrip("/api")
        S.login(host=T.API_HOST.rstrip("/api"), web_host=T.WEB_HOST)
        assert os.environ[SwanLabEnv.API_HOST.value] == T.API_HOST
        assert os.environ[SwanLabEnv.WEB_HOST.value] == T.WEB_HOST


@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
class TestInitExpByEnv:
    """
    测试通过环境变量创建实验
    """

    def test_workspace(self):
        """
        测试通过环境变量创建实验
        """
        os.environ[SwanLabEnv.WORKSPACE.value] = generate()
        # 随便生成的一个workspace会报错
        with pytest.raises(ValueError) as e:
            S.init()
        assert str(e.value) == 'Space `{}` not found'.format(os.environ[SwanLabEnv.WORKSPACE.value])
        del os.environ[SwanLabEnv.WORKSPACE.value]
        reset_client()

    def test_exp_name(self):
        """
        测试通过环境变量创建实验
        """
        exp_name = generate()
        os.environ[SwanLabEnv.EXP_NAME.value] = exp_name
        run = S.init()
        assert run.public.cloud.experiment_name == exp_name
        del os.environ[SwanLabEnv.EXP_NAME.value]

    def test_proj_name(self):
        """
        测试通过环境变量创建实验
        """
        proj_name = generate()
        os.environ[SwanLabEnv.PROJ_NAME.value] = proj_name
        run = S.init()
        assert run.public.cloud.project_name == proj_name
        del os.environ[SwanLabEnv.PROJ_NAME.value]

    def test_env_params_priority(self):
        """
        环境变量的优先级低于函数参数
        """
        exp_name = generate()
        os.environ[SwanLabEnv.EXP_NAME.value] = exp_name
        param_exp_name = generate()
        run = S.init(experiment_name=param_exp_name)
        assert run.public.cloud.experiment_name == param_exp_name


@pytest.mark.timeout(10)
@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
def test_init_again_after_terminal_login(monkeypatch):
    """
    测试实验结束后再次开启新实验，此时不需要再次登录
    """
    api_key = os.environ[SwanLabEnv.API_KEY.value]
    del os.environ[SwanLabEnv.API_KEY.value]
    num = 0

    def choose_mode(_):
        nonlocal num
        if num == 0:
            num += 1
            return '1'
        else:
            raise RuntimeError("This function should not be called!")

    def input_api_key(_):
        nonlocal num
        if num == 1:
            num += 1
            return api_key
        else:
            raise RuntimeError("This function should not be called!")

    monkeypatch.setattr("builtins.input", choose_mode)
    monkeypatch.setattr("getpass.getpass", input_api_key)

    S.init()
    S.finish()

    def raise_error(_):
        raise RuntimeError("This function should not be called!")

    monkeypatch.setattr("builtins.input", raise_error)
    S.init()


@pytest.mark.timeout(10)
@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
def test_init_again_after_code_login(monkeypatch):
    """
    测试在代码登录的情况下实验结束后再次开启新实验，此时不需要再次登录
    """

    def raise_error(_):
        raise RuntimeError("This function should not be called!")

    monkeypatch.setattr("builtins.input", raise_error)
    api_key = os.environ[SwanLabEnv.API_KEY.value]
    assert not os.path.exists(os.environ[SwanLabEnv.SWANLAB_FOLDER.value]), "SwanLab folder should be empty before test"

    del os.environ[SwanLabEnv.API_KEY.value]
    S.login(api_key=api_key)
    assert not os.path.exists(os.environ[SwanLabEnv.SWANLAB_FOLDER.value]), "SwanLab folder should be empty after login"

    S.init()
    S.finish()
    with pytest.raises(ValueError):
        get_client()
    assert os.getenv(SwanLabEnv.API_KEY.value) == api_key
    S.init()
    assert not os.path.exists(
        os.environ[SwanLabEnv.SWANLAB_FOLDER.value]
    ), "SwanLab folder should be empty after re-init"


@pytest.mark.parametrize("config", [1, [], (), True])
def test_init_error_config(config):
    """
    测试传入不同的config参数时，是否抛出错误对应错误
    """
    with pytest.raises(TypeError):
        swanlab.init(config=config, mode="disabled")


def test_init_error_config_name():
    """
    测试传入 string 类型参数
    此时会抛出 ValueError，因为不是指定的配置文件名后缀
    """
    with pytest.raises(ValueError):
        swanlab.init(mode="disabled", config="test_config")


@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
class TestResume:
    """
    测试resume功能
    """

    @pytest.mark.parametrize("resume", ["must", "allow"])
    @pytest.mark.parametrize("mode", ["local", "disabled", "offline"])
    def test_allow_resume_with_invalid_mode(self, mode, resume):
        """
        除了 cloud 以外的模式不允许 resume='allow'
        """
        with pytest.raises(RuntimeError) as e:
            S.init(resume='allow', mode=mode)
        assert str(e.value) == "You can only use resume in cloud mode."

    @pytest.mark.parametrize("resume", ["must", "allow"])
    def test_sync_step(self, resume):
        """
        测试 resume 时同步 step 参数
        """
        run = S.init()
        run.log({"loss": 0.3, "accuracy": 0.7}, step=1)
        run.finish()
        import time

        time.sleep(5)
        new_run = S.init(resume=resume, id=run.id, reinit=True)
        assert new_run.id == run.id, "New run ID should be the same as the previous one"
        # 上传相同 step 的指标，此时报错
        ll = new_run.log({"loss": 0.3, "accuracy": 0.7}, step=1)
        assert ll["loss"].is_error
        assert ll["accuracy"].is_error
        # 上传大于 step=1 的指标，此时不会报错
        ll = new_run.log({"loss": 0.5, "accuracy": 0.8})
        assert not ll["loss"].is_error, "Expected loss metric to be logged successfully after reinit"
        assert not ll["accuracy"].is_error, "Expected accuracy metric to be logged successfully after reinit"
        assert ll["loss"].metric_step == 2, "Expected loss metric step to be 2 after reinit"
        assert ll["accuracy"].metric_step == 2, "Expected accuracy metric step to be 2 after reinit"

    # ---------------------------------- never 部分 ----------------------------------
    def test_never_resume(self):
        """
        测试never模式下的resume功能
        """
        run = S.init(resume='never')
        run_store = get_run_store()
        assert run.id is not None, "Run ID should be generated"
        assert run_store.run_id == run.id, "Run store ID should match the run ID"
        run.log({"loss": 0.1, "accuracy": 0.9}, step=1)
        run.finish()
        # 再次初始化，应该是一个新的实验
        new_run = S.init(resume='never')
        assert new_run.id != run.id, "New run ID should be different from the previous one"

    def test_never_with_id(self):
        """
        never 模式不允许传递 id 参数
        """
        with pytest.raises(RuntimeError) as e:
            S.init(resume='never', id='test_id')
        assert str(e.value) == "You can't pass id when resume=never or resume=False."

    # ---------------------------------- allow 部分 ----------------------------------

    def test_allow_resume(self):
        """
        测试allow模式下的resume功能
        """
        run = S.init(resume='allow')
        run_store = get_run_store()
        assert run.id is not None, "Run ID should be generated"
        assert run_store.run_id == run.id, "Run store ID should match the run ID"
        run.log({"loss": 0.1, "accuracy": 0.9}, step=1)
        run.finish()
        import time

        time.sleep(5)
        # 再次初始化，应该是一个新的实验
        new_run = S.init(resume='allow', id=run.id)
        assert new_run.id == run.id, "New run ID should be the same as the previous one"
        # 校验 run_store ID 是否在云端实验连接中
        assert run.id in new_run.public.cloud.experiment_url, "New run should be linked to the previous run"

    def test_allow_with_id(self):
        """
        allow 模式允许传递 id 参数，此时根据此 id 初始化实验
        """
        id = "".join(random.choices("abcdefghijklmnopqrstuvwxyz0123456789", k=21))
        run = S.init(resume='allow', id=id)
        run_store = get_run_store()
        assert run.id == id, "Run ID should match the provided ID"
        assert run_store.run_id == id, "Run store ID should match the provided ID"
        # 再次初始化实验
        new_run = S.init(resume='allow', id=id)
        assert new_run.id == id, "New run ID should be the same as the provided ID"

    # ---------------------------------- must 部分 ----------------------------------

    def test_must_resume(self):
        """
        must 模式下之前的实验必须存在，否则会报错
        """
        run = S.init()
        run.finish()
        # 再次初始化，应该是一个新的实验
        new_run = S.init(resume='must', id=run.id)
        assert new_run.id == run.id, "New run ID should be the same as the previous one"
        # 校验 run_store ID 是否在云端实验连接中
        assert run.id in new_run.public.cloud.experiment_url, "New run should be linked to the previous run"

    def test_must_no_id(self):
        """
        在 must 模式下必须传递 id 参数，否则会报错
        """
        with pytest.raises(ValueError) as e:
            S.init(resume='must')
        assert str(e.value) == "You must pass id when resume=must."


@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
class TestResumeWithEnv:
    @staticmethod
    def teardown_method():
        """
        清理环境变量
        """
        if SwanLabEnv.RESUME.value in os.environ:
            del os.environ[SwanLabEnv.RESUME.value]

    def test_resume_env(self):
        """
        测试通过环境变量设置 resume 参数
        """
        os.environ[SwanLabEnv.RESUME.value] = "allow"
        S.init()
        run_store = get_run_store()
        assert run_store.resume == "allow"

    def test_resume_env_passed(self):
        """
        显式传递 resume 参数时，环境变量中的 resume 参数不生效
        """
        os.environ[SwanLabEnv.RESUME.value] = "allow"
        S.init(resume="never")
        run_store = get_run_store()
        assert run_store.resume == "never"


@pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
def test_online():
    """
    测试在线模式
    """
    run = S.init(mode="online")
    assert run.mode == "cloud"
    run.finish()
